# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import raise_if_err, \
    VNXMirrorException, VNXMirrorImageNotFoundError
from storops.lib.common import check_text, instance_cache
from storops.vnx.enums import VNXMirrorViewRecoveryPolicy
from storops.vnx.enums import VNXMirrorViewSyncRate
import storops.vnx.resource.lun
from storops.vnx.resource import VNXCliResource, VNXCliResourceList

__author__ = 'Cedric Zhuang'


class VNXMirrorViewImage(VNXCliResource):
    @staticmethod
    def get_id(image):
        if isinstance(image, VNXMirrorViewImage):
            image = image.uid
        try:
            image = check_text(image)
        except ValueError:
            raise ValueError('invalid image id supplied: {}'
                             .format(image))
        return image

    @property
    def wwn(self):
        return self.uid


class VNXMirrorViewImageList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorViewImage


class VNXMirrorView(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXMirrorView, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_view(name=self._name, poll=self.poll)

    @classmethod
    def create(cls, cli, name, src_lun, use_write_intent_log=True):
        lun_clz = storops.vnx.resource.lun.VNXLun
        lun_id = lun_clz.get_id(src_lun)
        out = cli.create_mirror_view(name, lun_id, use_write_intent_log)
        raise_if_err(out, default=VNXMirrorException)
        return VNXMirrorView(name, cli=cli)

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXMirrorViewList(cli)
        else:
            ret = VNXMirrorView(name, cli)
        return ret

    def add_image(self, sp_ip, lun_id,
                  recovery_policy=VNXMirrorViewRecoveryPolicy.AUTO,
                  sync_rate=VNXMirrorViewSyncRate.HIGH):
        if hasattr(sp_ip, 'spa_ip'):
            sp_ip = sp_ip.spa_ip

        lun_clz = storops.vnx.resource.lun.VNXLun
        lun_id = lun_clz.get_id(lun_id)
        out = self._cli.add_mirror_view_image(self._get_name(), sp_ip, lun_id,
                                              recovery_policy, sync_rate,
                                              poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def get_image(self, image_id):
        for image in self.images:
            if image.uid == image_id:
                ret = image
                break
        else:
            raise VNXMirrorImageNotFoundError(
                'image {} not found in mirror view {}.'.format(
                    image_id, self._get_name()))
        return ret

    @staticmethod
    def _get_image_id(image_id):
        return VNXMirrorViewImage.get_id(image_id)

    @property
    @instance_cache
    def primary_image(self):
        for image in self.images:
            if image.is_primary:
                ret = image
                break
        else:
            ret = None
        return ret

    @property
    @instance_cache
    def secondary_image(self):
        for image in self.images:
            if not image.is_primary:
                ret = image
                break
        else:
            ret = None
        return ret

    @property
    def is_primary(self):
        return self.remote_mirror_status == 'Mirrored'

    @property
    def primary_image_id(self):
        return self.primary_image.uid

    @property
    def secondary_image_id(self):
        image = self.secondary_image
        if image is None:
            raise VNXMirrorImageNotFoundError(
                'no secondary image exists for this mirror view.')
        return image.uid

    def remove_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.delete_mirror_view_image(self._get_name(), image_id,
                                                 poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def fracture_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.mirror_view_fracture_image(self._get_name(), image_id,
                                                   poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def sync_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.mirror_view_sync_image(self._get_name(), image_id,
                                               poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def promote_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.mirror_view_promote_image(self._get_name(), image_id,
                                                  poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def delete(self, force=False):
        if force:
            if self.secondary_image:
                self.remove_image()

        out = self._cli.delete_mirror_view(self._get_name())
        raise_if_err(out, default=VNXMirrorException)


class VNXMirrorViewList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorView

    def __init__(self, cli=None, src_lun=None, tgt_lun=None):
        super(VNXMirrorViewList, self).__init__()
        self._cli = cli
        self._src_lun = src_lun
        self._tgt_lun = tgt_lun

    def _filter(self, item):
        if self._src_lun is None and self._tgt_lun is None:
            ret = True
        else:
            ret = False
            pi = item.primary_image
            si = item.secondary_image
            if self._src_lun is not None:
                ret |= self._src_lun.wwn == pi.logical_unit_uid
            if self._tgt_lun is not None and si is not None:
                ret |= self._tgt_lun.wwn == si.logical_unit_uid
        return ret

    def _get_raw_resource(self):
        return self._cli.get_mirror_view(poll=self.poll)


class VNXMirrorGroup(VNXCliResource):
    def __init__(self, cli, name=None):
        super(VNXMirrorGroup, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_group(name=self._name, poll=self.poll)

    @classmethod
    def create(cls, cli, name, mirror=None):
        out = cli.create_mirror_group(name)
        raise_if_err(out, default=VNXMirrorException)
        group = VNXMirrorGroup(cli=cli, name=name)
        if mirror is not None:
            # Add to mirror group as well
            group.add_mirror(mirror)
        return group

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXMirrorGroupList(cli)
        else:
            ret = VNXMirrorGroup(cli, name)
        return ret

    def add_mirror(self, mirror):
        mirror_name = mirror._get_name()
        out = self._cli.add_to_mirror_group(self._get_name(), mirror_name,
                                            poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def remove_mirror(self, mirror):
        mirror_name = mirror._get_name()
        out = self._cli.remove_from_mirror_group(self._get_name(), mirror_name,
                                                 poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def fracture_group(self):
        out = self._cli.fracture_mirror_group(self._get_name(), poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def sync_group(self):
        out = self._cli.sync_mirror_group(self._get_name(), poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def promote_group(self):
        out = self._cli.promote_mirror_group(self._get_name(), poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def delete(self, force=False):
        out = self._cli.delete_mirror_group(self._get_name(), force=force)
        raise_if_err(out, default=VNXMirrorException)

    @property
    def mirrors(self):
        mirrors = []
        for m in self.group_mirrors:
            mirrors.append(VNXMirrorView(cli=self._cli, name=m.mirror_name))
        return mirrors


class VNXMirrorGroupMirror(VNXCliResource):
    def __init__(self, cli=None, name=None):
        super(VNXMirrorGroupMirror, self).__init__()
        self._cli = cli
        self._name = name


class VNXMirrorGroupMirrorList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorGroupMirror

    def __init__(self, cli=None):
        super(VNXMirrorGroupMirrorList, self).__init__()
        self._cli = cli


class VNXMirrorGroupList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorGroup

    def __init__(self, cli=None, name=None):
        super(VNXMirrorGroupList, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_group(name=self._name, poll=self.poll)


class VNXMirrorViewAsync(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXMirrorViewAsync, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_view_async(name=self._name, poll=self.poll)

    @classmethod
    def create(cls, cli, name, src_lun):
        lun_clz = storops.vnx.resource.lun.VNXLun
        lun_id = lun_clz.get_id(src_lun)
        out = cli.create_mirror_view_async(name, lun_id)
        raise_if_err(out, default=VNXMirrorException)
        return VNXMirrorViewAsync(name, cli=cli)

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXMirrorViewAsyncList(cli)
        else:
            ret = VNXMirrorViewAsync(name, cli)
        return ret

    def add_image(self, sp_ip, lun_id,
                  recovery_policy=VNXMirrorViewRecoveryPolicy.AUTO,
                  sync_rate=VNXMirrorViewSyncRate.HIGH):
        if hasattr(sp_ip, 'spa_ip'):
            sp_ip = sp_ip.spa_ip

        lun_clz = storops.vnx.resource.lun.VNXLun
        lun_id = lun_clz.get_id(lun_id)
        out = self._cli.add_mirror_view_image_async(self._get_name(),
                                                    sp_ip, lun_id,
                                                    recovery_policy,
                                                    sync_rate,
                                                    poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def get_image(self, image_id):
        for image in self.images:
            if image.uid == image_id:
                ret = image
                break
        else:
            raise VNXMirrorImageNotFoundError(
                'image {} not found in mirror view {}.'.format(
                    image_id, self._get_name()))
        return ret

    @staticmethod
    def _get_image_id(image_id):
        return VNXMirrorViewImage.get_id(image_id)

    @property
    @instance_cache
    def primary_image(self):
        for image in self.images:
            if image.is_primary:
                ret = image
                break
        else:
            ret = None
        return ret

    @property
    @instance_cache
    def secondary_image(self):
        for image in self.images:
            if not image.is_primary:
                ret = image
                break
        else:
            ret = None
        return ret

    @property
    def is_primary(self):
        return self.remote_mirror_status == 'Mirrored'

    @property
    def primary_image_id(self):
        return self.primary_image.uid

    @property
    def secondary_image_id(self):
        image = self.secondary_image
        if image is None:
            raise VNXMirrorImageNotFoundError(
                'no secondary image exists for this mirror view.')
        return image.uid

    def remove_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.delete_mirror_view_async_image(self._get_name(),
                                                       image_id,
                                                       poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def fracture_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.mirror_view_async_fracture_image(self._get_name(),
                                                         image_id,
                                                         poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def sync_image(self, image_id=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        out = self._cli.mirror_view_async_sync_image(self._get_name(),
                                                     image_id,
                                                     poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def promote_image(self, image_id=None, promote_type=None):
        if image_id is None:
            image_id = self.secondary_image_id

        image_id = self._get_image_id(image_id)
        c = self._cli
        out = c.mirror_view_async_promote_image(self._get_name(),
                                                image_id,
                                                poll=self.poll,
                                                promote_type=promote_type)
        raise_if_err(out, default=VNXMirrorException)

    def delete(self, force=False):
        if force:
            if self.secondary_image:
                self.remove_image()

        out = self._cli.delete_mirror_view_async(self._get_name())
        raise_if_err(out, default=VNXMirrorException)


class VNXMirrorViewAsyncList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorViewAsync

    def __init__(self, cli=None, src_lun=None, tgt_lun=None):
        super(VNXMirrorViewAsyncList, self).__init__()
        self._cli = cli
        self._src_lun = src_lun
        self._tgt_lun = tgt_lun

    def _filter(self, item):
        if self._src_lun is None and self._tgt_lun is None:
            ret = True
        else:
            ret = False
            pi = item.primary_image
            si = item.secondary_image
            if self._src_lun is not None:
                ret |= self._src_lun.wwn == pi.logical_unit_uid
            if self._tgt_lun is not None and si is not None:
                ret |= self._tgt_lun.wwn == si.logical_unit_uid
        return ret

    def _get_raw_resource(self):
        return self._cli.get_mirror_view_async(poll=self.poll)


class VNXMirrorGroupAsync(VNXCliResource):
    def __init__(self, cli, name=None):
        super(VNXMirrorGroupAsync, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_group_async(name=self._name,
                                                poll=self.poll)

    @classmethod
    def create(cls, cli, name, mirror=None):
        out = cli.create_mirror_group_async(name)
        raise_if_err(out, default=VNXMirrorException)
        group = VNXMirrorGroupAsync(cli=cli, name=name)
        if mirror is not None:
            # Add to mirror group as well
            group.add_mirror(mirror)
        return group

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXMirrorGroupAsyncList(cli)
        else:
            ret = VNXMirrorGroupAsync(cli, name)
        return ret

    def add_mirror(self, mirror):
        mirror_name = mirror._get_name()
        out = self._cli.add_to_mirror_group_async(self._get_name(),
                                                  mirror_name,
                                                  poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def remove_mirror(self, mirror):
        mirror_name = mirror._get_name()
        out = self._cli.remove_from_mirror_group_async(self._get_name(),
                                                       mirror_name,
                                                       poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def fracture_group(self):
        out = self._cli.fracture_mirror_group_async(self._get_name(),
                                                    poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def sync_group(self):
        out = self._cli.sync_mirror_group_async(self._get_name(),
                                                poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def promote_group(self, promote_type=None):
        out = self._cli.promote_mirror_group_async(self._get_name(),
                                                   promote_type=promote_type,
                                                   poll=self.poll)
        raise_if_err(out, default=VNXMirrorException)

    def delete(self, force=False):
        out = self._cli.delete_mirror_group_async(self._get_name(),
                                                  force=force)
        raise_if_err(out, default=VNXMirrorException)

    @property
    def mirrors(self):
        mirrors = []
        for m in self.group_mirrors:
            mirrors.append(VNXMirrorViewAsync(cli=self._cli,
                                              name=m.mirror_name))
        return mirrors


class VNXMirrorGroupMirrorAsync(VNXCliResource):
    def __init__(self, cli=None, name=None):
        super(VNXMirrorGroupMirrorAsync, self).__init__()
        self._cli = cli
        self._name = name


class VNXMirrorGroupMirrorAsyncList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorGroupMirrorAsync

    def __init__(self, cli=None):
        super(VNXMirrorGroupMirrorAsyncList, self).__init__()
        self._cli = cli


class VNXMirrorGroupAsyncList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXMirrorGroupAsync

    def __init__(self, cli=None, name=None):
        super(VNXMirrorGroupAsyncList, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_mirror_group_async(name=self._name,
                                                poll=self.poll)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import codecs
import json
import logging
import math
import os
from os.path import dirname, abspath, join, exists, basename

import fasteners
from hamcrest.core.base_matcher import BaseMatcher

from storops.lib.common import instance_cache, get_lock_file, get_data_file

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


def read_test_file(folder, name):
    p = dirname(abspath(__file__))
    p = join(p, folder)
    p = join(p, name)
    if not exists(p):
        log.warn('cannot find mock output file: %s, '
                 'default to empty string.', p)
        raise IOError('file: {} not found.'.format(p))
    else:
        with codecs.open(p, 'r', 'utf-8') as f:
            if basename(p) not in ('index.xml', 'index.json'):
                log.info('read mock file: %s', p)
            ret = f.read()

    return ret


class ConnectorMock(object):
    base_folder = ''

    def __init__(self, output=None, mock_map=None):
        self._output = output
        self._mock_map = mock_map

    @classmethod
    def read_file(cls, filename, *sub_folders):
        folder = join(cls.base_folder, *sub_folders)
        return read_test_file(folder, filename)

    def update_mock_output(self, output, mock_map):
        self._output = output
        self._mock_map = mock_map

    def get_file_in_mock_map(self, param):
        ret = None
        if self._mock_map is not None:
            for key in self._mock_map.keys():
                if param.startswith(key):
                    ret = self._mock_map[key]
                    break
        return ret

    def get_mock_output(self, inputs):
        filename = self.get_filename(inputs)
        folder = self.get_folder(inputs)
        from_map = self.get_file_in_mock_map(filename)
        if self._output is not None:
            filename = self._output
        elif from_map:
            filename = from_map

        return read_test_file(folder, filename)

    def get_filename(self, inputs):
        raise NotImplementedError('specify the inputs-filename mapping.')

    def get_folder(self, inputs):
        raise NotImplementedError('mock data folder not specified.')


class PersistedDict(object):
    def __init__(self, name=None, default=None, folder=None):
        if name is None:
            name = self.__hash__()
        self._name = name
        self._default = default

    @property
    def lock_file_name(self):
        return get_lock_file('{}.lck'.format(self._name))

    @property
    @instance_cache
    def data_file_name(self):
        return get_data_file('{}.json'.format(self._name))

    @property
    @instance_cache
    def lock(self):
        return fasteners.InterProcessLock(self.lock_file_name)

    @property
    def dict(self):
        if exists(self.data_file_name):
            with open(self.data_file_name) as f:
                try:
                    ret = json.load(f)
                except ValueError:
                    ret = {}
        else:
            ret = {}
        return ret

    def __setitem__(self, key, value):
        with self.lock:
            data = self.dict
            data[key] = value
            with open(self.data_file_name, 'w') as f:
                if log.isEnabledFor(level=logging.DEBUG):
                    s = json.dumps(data, indent=4, sort_keys=True)
                    log.debug('set dict: {}\n{}'
                              .format(self.data_file_name, s))
                json.dump(data, f, indent=4, sort_keys=True)

    def __getitem__(self, item):
        with self.lock:
            d = self.dict
            if item in d:
                ret = d[item]
            elif self._default:
                if callable(self._default):
                    ret = self._default()
                else:
                    ret = self._default
            else:
                ret = d[item]
        return ret

    def __len__(self):
        return len(self.dict)

    def clear(self):
        with self.lock:
            with open(self.data_file_name, 'w') as f:
                json.dump({}, f)

    def destroy(self):
        s = json.dumps(self.dict, indent=4, sort_keys=True)
        log.debug('destroy dict {}: \n{}'.format(self.data_file_name, s))
        if exists(self.data_file_name):
            os.remove(self.data_file_name)

    def clear_lock_file(self):
        if exists(self.lock_file_name):
            os.remove(self.lock_file_name)


class IsNaN(BaseMatcher):
    def __init__(self):
        self.value = None

    def _matches(self, value):
        self.value = value
        return math.isnan(value)

    def describe_to(self, description):
        description.append_text('"{}" should be NaN.'.format(self.value))


def is_nan():
    return IsNaN()



import functools
from unittest import TestCase

import ddt
from hamcrest import assert_that, equal_to, raises

from storops.exception import SystemAPINotSupported
from storops.lib.resource import Resource
from storops.lib.version import version, Criteria
from storops_test.unity.rest_mock import t_rest, patch_rest
from storops_test.vnx.cli_mock import patch_cli, t_cli


class ParentResource(Resource):
    @classmethod
    def parent_cls_method(cls, cli, o):
        return o

    @staticmethod
    def parent_static_method(cli, o):
        return o


@version('>2')
class DemoResource1(ParentResource):
    def __init__(self, cli):
        super(DemoResource1, self).__init__()
        self._cli = cli

    @classmethod
    def cls_method(cls, cli, o):
        return o

    @classmethod
    def cls_method_wo_param(cls):
        return True

    @staticmethod
    def static_method(cli, o):
        return o

    @staticmethod
    def static_method_wo_param():
        return True

    @classmethod
    def _get_parser(cls):
        pass

    def method(self, o):
        return o

    @property
    def system_version(self):
        return self._cli.system_version


class RequireVersionDecorateClassTest(TestCase):
    @patch_cli
    def test_cls_method_in_parent_support(self):
        self._test_cls_or_static_method_support(t_cli('3'),
                                                'parent_cls_method')

    @patch_rest
    def test_cls_method_in_parent_unsupport(self):
        self._test_cls_or_static_method_unsupport(t_rest('2'),
                                                  'parent_cls_method')

    @patch_rest
    def test_static_method_in_parent_support(self):
        self._test_cls_or_static_method_support(t_rest('3'),
                                                'parent_static_method')

    @patch_cli
    def test_static_method_in_parent_unsupport(self):
        self._test_cls_or_static_method_unsupport(t_cli('2'),
                                                  'parent_static_method')

    @patch_cli
    def test_cls_method_support(self):
        self._test_cls_or_static_method_support(t_cli('3'),
                                                'cls_method')

    @patch_rest
    def test_cls_method_unsupport(self):
        self._test_cls_or_static_method_unsupport(t_rest('2'),
                                                  'cls_method')

    @patch_rest
    def test_static_method_support(self):
        self._test_cls_or_static_method_support(t_rest('3'),
                                                'static_method')

    @patch_cli
    def test_static_method_unsupport(self):
        self._test_cls_or_static_method_unsupport(t_cli('2'),
                                                  'static_method')

    def test_static_method_wo_param(self):
        assert_that(DemoResource1.static_method_wo_param(), equal_to(True))

    def test_cls_method_wo_param(self):
        assert_that(DemoResource1.cls_method_wo_param(), equal_to(True))

    @patch_cli
    def test_new_instance_support(self):
        a = DemoResource1(t_cli('3'))
        assert_that(a.method('a'), equal_to('a'))

    @patch_rest
    def test_new_instance_unsupport(self):
        new_obj = functools.partial(DemoResource1, t_rest('2'))
        assert_that(new_obj, raises(SystemAPINotSupported))

    def _test_cls_or_static_method_support(self, cli, method):
        assert_that(getattr(DemoResource1, method)(cli, 'a'), 'a')

    def _test_cls_or_static_method_unsupport(self, cli, method):
        call = functools.partial(getattr(DemoResource1, method), cli, 'a')
        assert_that(call, raises(SystemAPINotSupported))


class DemoResource2(Resource):
    def __init__(self, version):
        super(DemoResource2, self).__init__()
        self.version = version

    @classmethod
    @version('>2')
    def cls_method(cls, cli, o=None):
        return o

    @staticmethod
    @version('>2')
    def static_method(cli, o=None):
        return o

    @version('>2')
    def method(cls):
        return True

    @version('>2')
    def versioned_method(self):
        return '>2'

    @version('<2')  # noqa
    def versioned_method(self):
        return '<2'

    @property
    def system_version(self):
        return self.version


class RequireVersionDecorateMethodTest(TestCase):
    def test_decorate_method(self):
        res = DemoResource2('3')
        re = res.method()
        assert_that(re, equal_to(True))
        res = DemoResource2('1')
        assert_that(res.method, raises(SystemAPINotSupported))

    @patch_cli
    def test_decorate_cls_method_support(self):
        assert_that(DemoResource2.cls_method(t_cli('3'), 'b'), equal_to('b'))

    @patch_cli
    def test_decorate_cls_method_unsupport(self):
        cls_method = functools.partial(DemoResource2.cls_method, t_cli('2'))
        assert_that(cls_method, raises(SystemAPINotSupported))

    @patch_cli
    def test_decorate_static_method_support(self):
        assert_that(DemoResource2.static_method(t_cli('3'), 'b'),
                    equal_to('b'))

    @patch_cli
    def test_decorate_static_method_unsupport(self):
        static_method = functools.partial(DemoResource2.static_method,
                                          t_cli('2'))
        assert_that(static_method, raises(SystemAPINotSupported))

    def test_versioned_method(self):
        res = DemoResource2('3')
        re = res.versioned_method()
        assert_that(re, equal_to('>2'))
        res = DemoResource2('1')
        re = res.versioned_method()
        assert_that(re, equal_to('<2'))
        res = DemoResource2('2')
        assert_that(res.versioned_method, raises(SystemAPINotSupported))


@ddt.ddt
class CriteriaTest(TestCase):
    @ddt.data({'version': '5.0.1', 'criteria': '>5.0', 'result': True},
              {'version': None, 'criteria': '>5.0', 'result': True},
              {'version': '1.0', 'criteria': '>=1.0', 'result': True},
              {'version': '1.0.1', 'criteria': '>=1.0', 'result': True},
              {'version': '2.0.2.1', 'criteria': '<2.0.3', 'result': True},
              {'version': '2.1.a', 'criteria': '<2.0.3', 'result': False},
              {'version': '2.0.3', 'criteria': '<=2.0.3', 'result': True},
              {'version': '2.1.a', 'criteria': '<=2.0.3', 'result': False},
              {'version': '5.0', 'criteria': '=5.0', 'result': True},
              {'version': '5.0', 'criteria': '==5.0', 'result': True},
              {'version': '5.0.1', 'criteria': '= 5.0', 'result': False},
              {'version': '5.0.1', 'criteria': '!= 5.0', 'result': True},
              {'version': '5.0', 'criteria': '!=5.0', 'result': False},
              {'version': '2.1.0.2', 'criteria': '2.0.5<>4.0.3', 'result':
                  True},
              {'version': '4.0.3.1', 'criteria': '2.0.5<>4.0.3',
               'result': False},
              {'version': '5.0.1', 'criteria': '2.0.5<>4.0.3',
               'result': False})
    @ddt.unpack
    def test_criteria_test(self, version, criteria, result):
        criteria = Criteria.parse(criteria)
        assert_that(criteria.test(version), equal_to(result))



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Yong Huang'

LOG = logging.getLogger(__name__)


class UnityMigrationObject(UnityResource):
    @classmethod
    def discover(cls, cli, remote_system):
        """
        Discover migrationObject instances on the remoteSystem.

        :param remote_system: Remote system from which objects need to
            be migrated.
        :return:
        """

        req_body = cli.make_body(remoteSystem=remote_system)
        resp = cli.type_action(cls().resource_class, 'discover', **req_body)
        resp.raise_if_err()
        return resp


class UnityMigrationObjectList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMigrationObject


class UnityLunInfo(UnityResource):
    pass


class UnityLunInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLunInfo


class UnityFilesystemInfo(UnityResource):
    pass


class UnityFilesystemInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityFilesystemInfo


class UnityInterfaceInfo(UnityResource):
    pass


class UnityInterfaceInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityInterfaceInfo


class UnityCifsServerInfo(UnityResource):
    pass


class UnityCifsServerInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityCifsServerInfo


class UnityVdmInfo(UnityResource):
    pass


class UnityVdmInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVdmInfo



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os
import unittest
from time import sleep

from hamcrest import assert_that, less_than, greater_than, none, equal_to, \
    has_items, contains_string

from storops.lib.common import get_data_file
from storops.lib.metric import PerfManager, MetricCounterRecords, MetricsDumper
from storops.unity.resource.disk import UnityDiskList, UnityDisk
from storops.unity.resource.lun import UnityLun, UnityLunList

__author__ = 'Cedric Zhuang'


class PerfManagerTest(unittest.TestCase):
    def perf_mon(self):
        return PerfManager()

    def test_enable_perf_metric(self):
        cli = self.perf_mon()
        assert_that(cli.is_perf_metric_enabled(), equal_to(False))
        cli.enable_perf_metric(0, lambda: 1)
        assert_that(cli.is_perf_metric_enabled(), equal_to(True))
        cli.disable_perf_metric()
        assert_that(cli.is_perf_metric_enabled(), equal_to(False))

    def test_double_enable(self):
        cli = self.perf_mon()
        c = []
        cli.enable_perf_metric(0.1, lambda: c.append(1))
        sleep(0.5)
        assert_that(len(c), greater_than(1))
        assert_that(len(c), less_than(8))
        cli.enable_perf_metric(10, lambda: c.append(1))
        sleep(0.5)
        assert_that(len(c), less_than(10))
        cli.disable_perf_metric()
        assert_that(cli.prev_counter, none())
        assert_that(cli.curr_counter, none())
        assert_that(cli.is_perf_metric_enabled(), equal_to(False))

    def test_is_perf_metric_enabled_rsc_default(self):
        cli = self.perf_mon()
        cli.enable_perf_metric(0, lambda: 1)
        enabled = cli.is_perf_metric_enabled(UnityDiskList(cli=cli))
        assert_that(enabled, equal_to(True))
        enabled = cli.is_perf_metric_enabled(UnityDisk(_id='', cli=cli))
        assert_that(enabled, equal_to(True))

    def test_is_perf_metric_enabled_rsc_specific(self):
        cli = self.perf_mon()
        cli.enable_perf_metric(0, lambda: 1, [UnityDisk])
        enabled = cli.is_perf_metric_enabled(UnityDiskList(cli=cli))
        assert_that(enabled, equal_to(True))

        enabled = cli.is_perf_metric_enabled(UnityLunList(cli=cli))
        assert_that(enabled, equal_to(False))

        enabled = cli.is_perf_metric_enabled(UnityDisk(_id='', cli=cli))
        assert_that(enabled, equal_to(True))

        enabled = cli.is_perf_metric_enabled(UnityLun(_id='', cli=cli))
        assert_that(enabled, equal_to(False))

    def test_is_perf_monitored_default(self):
        assert_that(self.perf_mon()._is_perf_monitored('abc'), equal_to(True))

    def test_is_perf_monitored_resource(self):
        cli = self.perf_mon()
        cli._rsc_clz_list = [UnityDisk]
        assert_that(cli._is_perf_monitored(UnityDisk('', cli=cli)),
                    equal_to(True))
        assert_that(cli._is_perf_monitored(UnityLun('', cli=cli)),
                    equal_to(False))

    def test_is_perf_monitored_resource_list(self):
        cli = self.perf_mon()
        cli._rsc_clz_list = [UnityDisk]
        assert_that(cli._is_perf_monitored(UnityDiskList(cli=cli)),
                    equal_to(True))
        assert_that(cli._is_perf_monitored(UnityLunList(cli=cli)),
                    equal_to(False))


class MetricCounterRecordsTest(unittest.TestCase):
    def test_max_count(self):
        records = MetricCounterRecords()
        records.add_results(1)
        records.add_results(2)
        records.add_results(3)
        assert_that(len(records), equal_to(2))
        assert_that(records._records, has_items(2, 3))
        assert_that(records.enabled, equal_to(True))

    def test_default_enabled(self):
        records = MetricCounterRecords()
        assert_that(records.enabled, equal_to(False))

    def test_reset(self):
        records = MetricCounterRecords()
        records.add_results(1)
        records.reset()
        assert_that(len(records), equal_to(0))
        assert_that(records.enabled, equal_to(False))

    def test_add_none_result(self):
        records = MetricCounterRecords()
        records.add_results(None)
        assert_that(len(records), equal_to(0))

    def test_curr_prev_value(self):
        records = MetricCounterRecords()
        assert_that(records.curr, none())
        assert_that(records.prev, none())

        records.add_results(1)
        assert_that(records.curr, equal_to(1))
        assert_that(records.prev, none())

        records.add_results(2)
        assert_that(records.curr, equal_to(2))
        assert_that(records.prev, equal_to(1))

        records.add_results(3)
        assert_that(records.curr, equal_to(3))
        assert_that(records.prev, equal_to(2))

        records.enabled = False
        assert_that(records.curr, none())
        assert_that(records.prev, none())


class SampleRscList(object):
    def __init__(self):
        self.list = [{'time': 1, 'name': 'a', 'ma': 1, 'mb': 2.0, 'mc': 'aaa'},
                     {'time': 1, 'name': 'b', 'ma': 4, 'mb': 5.0, 'mc': 'bbb'}]

    @staticmethod
    def metric_names():
        return ['ma', 'mb', 'mc']

    def __iter__(self):
        return self.list.__iter__()

    @staticmethod
    def _hdr_cb(rsc):
        return list(map(str, [rsc['time'], rsc['name']]))


class MetricsDumperTest(unittest.TestCase):
    dumper = MetricsDumper(SampleRscList(), ['time', 'name'],
                           SampleRscList._hdr_cb)
    filename = get_data_file('temp_test_metric.csv')

    @classmethod
    def tearDownClass(cls):
        # noinspection PyBroadException
        try:
            os.remove(cls.filename)
        except Exception:
            pass

    def test_metric_names(self):
        assert_that(self.dumper.metric_names, has_items('ma', 'mb', 'mc'))

    def test_get_metrics_csv_header(self):
        assert_that(self.dumper.get_metrics_csv_header(),
                    equal_to('time,name,ma,mb,mc'))

    def test_get_metrics_csv_data(self):
        data = self.dumper.get_metrics_csv_data()
        assert_that(data, contains_string('1,a,1,2.0,aaa'))
        assert_that(data, contains_string('1,b,4,5.0,bbb'))

    def test_data_line(self):
        assert_that(self.dumper.data_line(next(iter(SampleRscList()))),
                    has_items('1', 'a', '1', '2.0', 'aaa'))

    def test_persist_metric_data(self):
        filename = self.filename
        self.dumper.persist_metric_data(filename)
        exists = os.path.exists(filename)
        assert_that(exists, equal_to(True), '{} not found.'.format(filename))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from multiprocessing.pool import ThreadPool
from time import sleep
from unittest import TestCase

import bitmath
from hamcrest import assert_that, equal_to, close_to, only_contains, raises, \
    contains_string, has_items, not_none, none

from storops.exception import EnumValueNotFoundError
from storops.lib import common
from storops.lib.common import Dict, Enum, WeightedAverage, synchronized, \
    text_var, int_var, enum_var, yes_no_var, list_var, JsonPrinter, \
    get_lock_file, EnumList, round_3, RepeatedTimer, supplement_filesystem, \
    try_import
from storops.vnx.enums import VNXRaidType

log = logging.getLogger(__name__)


class DictTest(TestCase):
    def test_get_attr(self):
        result = Dict()
        result['a'] = 'A'
        assert_that(result.a, equal_to('A'))
        assert_that(result['a'], equal_to('A'))

    def test_get_attr_not_exists(self):
        def f():
            result = Dict()
            log.debug(result.a)

        assert_that(f, raises(AttributeError))


class SampleEnum(Enum):
    TYPE_A = 'type a'
    TYPE_B = 'type b'

    @classmethod
    def get_option_map(cls):
        return {
            None: [],
            cls.TYPE_A: '-a',
            cls.TYPE_B: '-b'
        }

    @classmethod
    def get_int_index(cls):
        return None, cls.TYPE_A, cls.TYPE_B


class SampleIntEnum(Enum):
    OK = 0
    NOT_FOUND = 1
    ERROR = 2


class SampleIntEnumList(EnumList):
    @classmethod
    def get_enum_class(cls):
        return SampleIntEnum


class EnumTest(TestCase):
    def test_verify_allow_none(self):
        # no error raised
        SampleIntEnum.verify(None)

    def test_verify_normal_value(self):
        # no error raised
        SampleIntEnum.verify(SampleIntEnum.NOT_FOUND)

    def test_verify_not_allow_none(self):
        def f():
            SampleIntEnum.verify(None, allow_none=False)

        assert_that(f, raises(ValueError, 'SampleIntEnum'))

    def test_verify_not_valid_value(self):
        def f():
            SampleIntEnum.verify(88)

        assert_that(f, raises(ValueError, 'SampleIntEnum'))

    def test_get_all(self):
        assert_that(len(SampleEnum.get_all()), equal_to(2))

    def test_get_opt(self):
        assert_that(SampleEnum.get_opt(SampleEnum.TYPE_A), equal_to('-a'))

    def test_from_int(self):
        assert_that(SampleEnum.from_int(2), equal_to(SampleEnum.TYPE_B))

    def test_from_int_not_found_in_index(self):
        def f():
            SampleEnum.from_int(10)

        assert_that(f, raises(EnumValueNotFoundError, 'for SampleEnum'))

    def test_from_int_enum(self):
        assert_that(SampleIntEnum.from_int(1),
                    equal_to(SampleIntEnum.NOT_FOUND))

    def test_from_int_not_found(self):
        def f():
            SampleIntEnum.from_int(10)

        assert_that(f, raises(EnumValueNotFoundError, 'for SampleIntEnum'))

    def test_parse_string(self):
        ret = SampleEnum.parse('type a')
        assert_that(ret, equal_to(SampleEnum.TYPE_A))

    def test_parse_invalid_str_enum(self):
        def f():
            SampleEnum.parse('n/a')

        assert_that(f, raises(EnumValueNotFoundError))

    def test_parse_invalid_int_enum(self):
        def f():
            SampleEnum.parse(999)

        assert_that(f, raises(EnumValueNotFoundError))
        assert_that(f, raises(ValueError))

    def test_invalid_value(self):
        def f():
            SampleEnum('type c')

        assert_that(f, raises(ValueError, 'not a valid SampleEnum'))

    def test_values(self):
        assert_that(SampleEnum.values(), only_contains('type a', 'type b'))

    def test_enum_list_parsing(self):
        assert_that(SampleIntEnumList.parse([1, 2]),
                    has_items(SampleIntEnum.NOT_FOUND, SampleIntEnum.ERROR))

    def test_enum_list_to_str(self):
        assert_that(str(SampleIntEnumList.parse([0, 1])),
                    equal_to('{"SampleIntEnumList": ["SampleIntEnum.OK", '
                             '"SampleIntEnum.NOT_FOUND"]}'))


class WeightedAverageTest(TestCase):
    def test_data_full(self):
        avg = WeightedAverage(3)
        avg.add(30, 24, 18, 12, 6)
        assert_that(avg.value(), equal_to(10))
        avg.add(30)
        assert_that(avg.value(), equal_to(19))

    def test_data_full_1(self):
        avg = WeightedAverage(6)
        avg.add(30, 24, 18, 12, 6, 6, 6)
        assert_that(avg.value(), close_to(8.85, 0.01))
        avg.add(30)
        assert_that(avg.value(), equal_to(14))

    def test_data_not_full(self):
        avg = WeightedAverage(3)
        avg.add(12, 6)
        assert_that(avg.value(), equal_to(8.4))

    def test_data_empty(self):
        avg = WeightedAverage(3)
        assert_that(avg.value(), equal_to(0))

    def test_size(self):
        avg = WeightedAverage(3)
        assert_that(avg.size, equal_to(3))
        avg.weight = [3]
        avg.add(4, 5, 7)
        assert_that(avg.size, equal_to(1))
        assert_that(avg.value(), equal_to(7))


class LockDemo(object):
    name = 1

    def __init__(self):
        self.call_count = 0

    @synchronized(name)
    def foo1(self):
        self.call_count += 1
        sleep(0.05)

    def foo2(self, value):
        @synchronized(value)
        def in_foo2():
            self.call_count += 1
            sleep(0.05)

        in_foo2()

    @synchronized()
    def bar(self, _):
        self.call_count += 1
        sleep(0.05)


class SynchronizedTest(TestCase):
    def test_synchronize_simple(self):
        demo = LockDemo()
        pool = ThreadPool(2)
        pool.imap(demo.bar, range(2))
        sleep(0.04)
        assert_that(demo.call_count, equal_to(1))
        sleep(0.05)
        assert_that(demo.call_count, equal_to(2))

    def test_synchronize_with_different_param(self):
        demo = LockDemo()
        pool = ThreadPool(2)
        pool.imap(demo.foo2, range(2))
        sleep(0.02)
        assert_that(demo.call_count, equal_to(2))

    def test_synchronize_with_same_param(self):
        demo = LockDemo()
        pool = ThreadPool(3)
        pool.imap(demo.foo2, (1, 1))
        pool.apply_async(demo.foo1)
        sleep(0.04)
        assert_that(demo.call_count, equal_to(1))
        sleep(0.05)
        assert_that(demo.call_count, equal_to(2))
        sleep(0.05)
        assert_that(demo.call_count, equal_to(3))


class VarTest(TestCase):
    def test_text_var(self):
        assert_that(text_var('-a', 'a'), only_contains('-a', 'a'))
        assert_that(text_var(None, 'a'), only_contains('a'))
        assert_that(text_var('-a', None), equal_to([]))

    def test_int_var(self):
        assert_that(int_var('-a', '1'), only_contains('-a', 1))
        assert_that(int_var(None, '1'), only_contains(1))
        assert_that(int_var('-a', None), equal_to([]))

    def test_enum_var(self):
        assert_that(enum_var('-a', 'r5', VNXRaidType),
                    only_contains('-a', 'r5'))
        assert_that(enum_var(None, 'r5', VNXRaidType), only_contains('r5'))
        assert_that(enum_var('-a', None, VNXRaidType), equal_to([]))

    def test_yes_no_var(self):
        assert_that(yes_no_var('-a', True), only_contains('-a', 'yes'))
        assert_that(yes_no_var('-a', False), only_contains('-a', 'no'))
        assert_that(yes_no_var('-a', None), equal_to([]))

    def test_list_var(self):
        assert_that(list_var('-a', []), equal_to([]))
        assert_that(list_var('-a', None), equal_to([]))
        assert_that(list_var('-a', ['b', 'c']), equal_to(['-a', 'b', 'c']))

        def _inner():
            list_var('-a', 'asf')

        assert_that(_inner, raises(ValueError))


class JsonPrinterDemo(JsonPrinter):
    def __init__(self):
        self.a = 1
        self.b = None

    def _get_properties(self, dec=0):
        return {'a': self.a, 'b': self.b}


class JsonPrinterTest(TestCase):
    def test_str_delete_null(self):
        j = JsonPrinterDemo()
        assert_that(str(j), equal_to('{"JsonPrinterDemo": {"a": 1}}'))


class CommonTest(TestCase):
    def test_get_lock_file(self):
        name = get_lock_file('a.lock')
        assert_that(name, contains_string('.storops'))
        assert_that(name, contains_string('a.lock'))

    def test_try_import_none(self):
        mod = try_import('paramiko_none')
        assert_that(mod, none())

    def test_try_import_normal(self):
        mod = try_import('os')
        assert_that(mod, not_none())


class RoundItTest(TestCase):
    def test_round3(self):
        @round_3
        def f():
            return 123.45678

        assert_that(f(), equal_to(123.457))


class RepeatedTimerTest(TestCase):
    def test_repeated_timer_normal(self):
        count = [0]
        results = []

        def f(v):
            count[0] += 1
            results.append(v * count[0])

        timer = RepeatedTimer(0.05, f, 2)
        timer.start()
        sleep(0.5)
        timer.stop()
        assert_that(results, has_items(2, 4, 6))

    def test_repeated_timer_daemon(self):
        results = [0]

        def f():
            results[0] += 1

        timer = RepeatedTimer(10, f)
        timer.start()
        assert_that(timer.is_daemon(), equal_to(True))
        assert_that(results[0], equal_to(0))


class SupplementFilesystemTest(TestCase):
    def test_size_larger_2(self):
        size_byte = bitmath.GiB(2.5).to_Byte().value
        new_size = supplement_filesystem(size_byte, True)
        assert_that(new_size, equal_to(bitmath.GiB(4).to_Byte().value))
        assert_that(str(new_size).isdigit(), equal_to(True))

    def test_size_equal_1(self):
        size_byte = bitmath.GiB(1).to_Byte().value
        new_size = supplement_filesystem(size_byte, True)
        assert_that(new_size, equal_to(bitmath.GiB(3).to_Byte().value))

    def test_user_cap_no(self):
        size_byte = bitmath.GiB(1).to_Byte().value
        new_size = supplement_filesystem(size_byte, False)
        assert_that(new_size, equal_to(bitmath.GiB(1).to_Byte().value))

        size_byte = bitmath.GiB(3).to_Byte().value
        new_size = supplement_filesystem(size_byte, False)
        assert_that(new_size, equal_to(bitmath.GiB(3).to_Byte().value))


class InitiatorFormatTest(TestCase):
    def test_iscsi_format_rh(self):
        assert_that(common.is_iscsi_uid('iqn.1994-05.com.redhat:4dc0b7e1edc9'),
                    equal_to(True))

    def test_iscsi_format_debian(self):
        assert_that(
            common.is_iscsi_uid('iqn.1993-08.org.debian:01:816ce05feaa6'),
            equal_to(True))

    def test_iscsi_format_invalid(self):
        assert_that(common.is_iscsi_uid('invalid format'), equal_to(False))

    def test_fc_format(self):
        assert_that(
            common.is_fc_uid(
                '20:00:00:90:DA:73:5C:D1:10:00:00:90:FB:54:4C:D1'),
            equal_to(True))

    def test_fc_format_invalid(self):
        assert_that(
            common.is_fc_uid('20:00:00:90:DA:73:5C:D1'), equal_to(False))


@common.singleton
class MockSingleton(object):
    pass


class SingletonTest(TestCase):
    def test_singleton(self):
        cls_a = MockSingleton()
        cls_b = MockSingleton()
        assert_that(id(cls_a), equal_to(id(cls_b)))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from hamcrest import equal_to, assert_that, raises, contains_string

from storops import cache
from storops.exception import VNXMirrorSyncImageError, \
    VNXMirrorPromotePrimaryError

from storops_comptest import vnx1, vnx2

__author__ = 'Cedric Zhuang'


@cache
def is_mirror_available():
    return (vnx1().is_mirror_view_sync_enabled() and
            vnx2().is_mirror_view_sync_enabled())


@pytest.mark.skipif(not is_mirror_available(),
                    reason='mirror view sync is not available.')
def test_create_mirror_view(multi_vnx_gf):
    mirror = multi_vnx_gf.mirror
    assert_that(mirror.existed, equal_to(True))


@pytest.mark.skipif(not is_mirror_available(),
                    reason='mirror view sync is not available.')
def test_add_image(multi_vnx_gf):
    mirror = multi_vnx_gf.mirror
    assert_that(len(mirror.images), equal_to(2))


@pytest.mark.skipif(not is_mirror_available(),
                    reason='mirror view sync is not available.')
def test_fracture_and_sync_image(multi_vnx_gf):
    mirror = multi_vnx_gf.mirror
    mirror.fracture_image(mirror.secondary_image)
    mirror.update()

    assert_that(mirror.secondary_image.condition, contains_string('fractured'))

    def sync():
        mirror.sync_image(mirror.images[0])

    sync()

    assert_that(sync, raises(VNXMirrorSyncImageError))


@pytest.mark.skipif(not is_mirror_available(),
                    reason='mirror view sync is not available.')
def test_promote_primary_image(multi_vnx_gf):
    def promote():
        mirror = multi_vnx_gf.mirror
        mirror.promote_image(mirror.primary_image)

    assert_that(promote, raises(VNXMirrorPromotePrimaryError))



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Veriion 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

import ddt
import hamcrest as hc

import storops
from storops.unity.resource import health, replication_session, \
    remote_system, remote_interface
from storops_test.unity import rest_mock as mock

__author__ = 'Ryan Liang'


@ddt.ddt
class UnityReplicationSessionTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(), _id='42949675124_FNM00150600267_0000'
                                   '_42949678638_FNM00152000052_0000')
        hc.assert_that(rs.current_transfer_est_remain_time, hc.equal_to(0))
        hc.assert_that(rs.daily_snap_replication_policy,
                       hc.instance_of(
                           replication_session.UnitySnapReplicationPolicy))
        hc.assert_that(rs.daily_snap_replication_policy.is_replicating_snaps,
                       hc.equal_to(False))
        hc.assert_that(rs.dst_resource_id, hc.equal_to('sv_3'))
        hc.assert_that(rs.dst_spa_interface,
                       hc.instance_of(remote_interface.UnityRemoteInterface))
        hc.assert_that(rs.dst_spa_interface.id,
                       hc.equal_to('FNM00152000052:if_7'))
        hc.assert_that(rs.dst_spb_interface,
                       hc.instance_of(remote_interface.UnityRemoteInterface))
        hc.assert_that(rs.dst_spb_interface.id,
                       hc.equal_to('FNM00152000052:if_6'))
        hc.assert_that(rs.dst_status,
                       hc.equal_to(storops.ReplicationSessionStatusEnum.OK))
        hc.assert_that(rs.health, hc.instance_of(health.UnityHealth))
        hc.assert_that(rs.health.value, hc.equal_to(storops.HealthEnum.OK))
        hc.assert_that(rs.hourly_snap_replication_policy,
                       hc.instance_of(
                           replication_session.UnitySnapReplicationPolicy))
        hc.assert_that(rs.hourly_snap_replication_policy.is_replicating_snaps,
                       hc.equal_to(False))
        hc.assert_that(rs.id, hc.equal_to(
            '42949675124_FNM00150600267_0000_42949678638_FNM00152000052_0000'))
        hc.assert_that(str(rs.last_sync_time),
                       hc.equal_to('2018-12-07 09:27:47+00:00'))
        hc.assert_that(rs.local_role, hc.equal_to(
            storops.ReplicationSessionReplicationRoleEnum.SOURCE))
        hc.assert_that(rs.max_time_out_of_sync, hc.equal_to(60))
        hc.assert_that(rs.members, hc.none())
        hc.assert_that(rs.name, hc.equal_to(
            'rep_sess_sv_1876_sv_3_FNM00150600267_FNM00152000052'))
        hc.assert_that(rs.network_status, hc.equal_to(
            storops.ReplicationSessionNetworkStatusEnum.OK))
        hc.assert_that(rs.remote_system,
                       hc.instance_of(remote_system.UnityRemoteSystem))
        hc.assert_that(rs.remote_system.id, hc.equal_to('RS_4'))
        hc.assert_that(rs.replication_resource_type, hc.equal_to(
            storops.ReplicationEndpointResourceTypeEnum.LUN))
        hc.assert_that(rs.src_resource_id, hc.equal_to('sv_1876'))
        hc.assert_that(rs.src_spa_interface,
                       hc.instance_of(remote_interface.UnityRemoteInterface))
        hc.assert_that(rs.src_spa_interface.id,
                       hc.equal_to('FNM00150600267:if_10'))
        hc.assert_that(rs.src_spb_interface,
                       hc.instance_of(remote_interface.UnityRemoteInterface))
        hc.assert_that(rs.src_spb_interface.id,
                       hc.equal_to('FNM00150600267:if_11'))
        hc.assert_that(rs.src_status,
                       hc.equal_to(storops.ReplicationSessionStatusEnum.OK))
        hc.assert_that(rs.status, hc.equal_to(
            storops.ReplicationOpStatusEnum.AUTO_SYNC_CONFIGURED))
        hc.assert_that(rs.sync_progress, hc.equal_to(0))
        hc.assert_that(rs.sync_state, hc.equal_to(
            storops.ReplicationSessionSyncStateEnum.IDLE))

    @mock.patch_rest
    def test_get_collection(self):
        rs_list = replication_session.UnityReplicationSessionList.get(
            mock.t_rest())
        hc.assert_that(len(rs_list), hc.equal_to(2))
        hc.assert_that(
            [rs.id for rs in rs_list],
            hc.equal_to(['42949675116_FNM00150600267_0000'
                         '_42949675117_FNM00150600267_0000',
                         '42949675124_FNM00150600267_0000'
                         '_42949678638_FNM00152000052_0000']))

    @mock.patch_rest
    @ddt.data(
        {'src_resource_id': 'sv_2498', 'dst_resource_id': 'sv_5',
         'max_time_out_of_sync': 60, 'name': 'remote-sv_2498-sv_5',
         'members': None, 'auto_initiate': False,
         'hourly_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=False),
         'daily_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=True, is_retention_same_as_source=False,
                 is_auto_delete=False, retention_duration=3600),
         'replicate_existing_snaps': False, 'remote_sys': 'RS_4',
         'src_spa_interface': 'FNM00152000052:if_7',
         'src_spb_interface': 'FNM00152000052:if_6',
         'dst_spa_interface': 'FNM00150600267:if_10',
         'dst_spb_interface': 'FNM00150600267:if_11'},
        {'src_resource_id': 'sv_2498', 'dst_resource_id': 'sv_5',
         'max_time_out_of_sync': 60, 'name': None,
         'members': None, 'auto_initiate': None,
         'hourly_snap_replication_policy': None,
         'daily_snap_replication_policy': None,
         'replicate_existing_snaps': None, 'remote_sys': None,
         'src_spa_interface': None, 'src_spb_interface': None,
         'dst_spa_interface': None, 'dst_spb_interface': None})
    @ddt.unpack
    def test_create(self, src_resource_id, dst_resource_id,
                    max_time_out_of_sync, name, members,
                    auto_initiate, hourly_snap_replication_policy,
                    daily_snap_replication_policy,
                    replicate_existing_snaps, remote_sys,
                    src_spa_interface, src_spb_interface,
                    dst_spa_interface, dst_spb_interface):

        if remote_sys:
            remote_sys = remote_system.UnityRemoteSystem(_id=remote_sys,
                                                         cli=mock.t_rest())
        if src_spa_interface:
            src_spa_interface = remote_interface.UnityRemoteInterface(
                _id=src_spa_interface, cli=mock.t_rest())
        if src_spb_interface:
            src_spb_interface = remote_interface.UnityRemoteInterface(
                _id=src_spb_interface, cli=mock.t_rest())
        if dst_spa_interface:
            dst_spa_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spa_interface, cli=mock.t_rest())
        if dst_spb_interface:
            dst_spb_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spb_interface, cli=mock.t_rest())
        rs = replication_session.UnityReplicationSession.create(
            mock.t_rest(), src_resource_id, dst_resource_id,
            max_time_out_of_sync, name=name, members=members,
            auto_initiate=auto_initiate,
            hourly_snap_replication_policy=hourly_snap_replication_policy,
            daily_snap_replication_policy=daily_snap_replication_policy,
            replicate_existing_snaps=replicate_existing_snaps,
            remote_system=remote_sys,
            src_spa_interface=src_spa_interface,
            src_spb_interface=src_spb_interface,
            dst_spa_interface=dst_spa_interface,
            dst_spb_interface=dst_spb_interface)
        hc.assert_that(rs.id, hc.equal_to(
            '42949675780_FNM00150600267_0000_42949678642_FNM00152000052_0000'))

    @mock.patch_rest
    @ddt.data(
        {'src_resource_id': 'sv_1876',
         'dst_resource_config':
             replication_session.UnityResourceConfig.to_embedded(
                 pool_id='pool_2', request_id='lun-rep-src3-liangr',
                 name='lun-rep-src3-liangr', is_thin_enabled=True,
                 size=10737418240,
                 tiering_policy=storops.TieringPolicyEnum.AUTOTIER_HIGH,
                 is_deduplication_enabled=False, is_compression_enabled=False),
         'max_time_out_of_sync': 60, 'name': 'remote-rep3',
         'remote_sys': 'RS_4',
         'src_spa_interface': 'FNM00152000052:if_7',
         'src_spb_interface': 'FNM00152000052:if_6',
         'dst_spa_interface': 'FNM00150600267:if_10',
         'dst_spb_interface': 'FNM00150600267:if_11',
         'dst_resource_element_configs': None, 'auto_initiate': False,
         'hourly_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=False),
         'daily_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=True, is_retention_same_as_source=False,
                 is_auto_delete=False, retention_duration=3600),
         'replicate_existing_snaps': False},
        {'src_resource_id': 'sv_1876',
         'dst_resource_config':
             replication_session.UnityResourceConfig.to_embedded(
                 pool_id='pool_2', request_id='lun-rep-src3-liangr',
                 name='lun-rep-src3-liangr', is_thin_enabled=True,
                 size=10737418240,
                 tiering_policy=storops.TieringPolicyEnum.AUTOTIER_HIGH,
                 is_deduplication_enabled=False, is_compression_enabled=False),
         'max_time_out_of_sync': 60, 'name': None, 'remote_sys': None,
         'src_spa_interface': None, 'src_spb_interface': None,
         'dst_spa_interface': None, 'dst_spb_interface': None,
         'dst_resource_element_configs': None, 'auto_initiate': None,
         'hourly_snap_replication_policy': None,
         'daily_snap_replication_policy': None,
         'replicate_existing_snaps': None})
    @ddt.unpack
    def test_create_with_dst_resource_provisioning(
            self, src_resource_id, dst_resource_config,
            max_time_out_of_sync, name=None, remote_sys=None,
            src_spa_interface=None, src_spb_interface=None,
            dst_spa_interface=None, dst_spb_interface=None,
            dst_resource_element_configs=None, auto_initiate=None,
            hourly_snap_replication_policy=None,
            daily_snap_replication_policy=None,
            replicate_existing_snaps=None):

        if remote_sys:
            remote_sys = remote_system.UnityRemoteSystem(_id=remote_sys,
                                                         cli=mock.t_rest())
        if src_spa_interface:
            src_spa_interface = remote_interface.UnityRemoteInterface(
                _id=src_spa_interface, cli=mock.t_rest())
        if src_spb_interface:
            src_spb_interface = remote_interface.UnityRemoteInterface(
                _id=src_spb_interface, cli=mock.t_rest())
        if dst_spa_interface:
            dst_spa_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spa_interface, cli=mock.t_rest())
        if dst_spb_interface:
            dst_spb_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spb_interface, cli=mock.t_rest())
        _cls = replication_session.UnityReplicationSession
        rs = _cls.create_with_dst_resource_provisioning(
            mock.t_rest(), src_resource_id, dst_resource_config,
            max_time_out_of_sync, name=name, remote_system=remote_sys,
            src_spa_interface=src_spa_interface,
            src_spb_interface=src_spb_interface,
            dst_spa_interface=dst_spa_interface,
            dst_spb_interface=dst_spb_interface,
            dst_resource_element_configs=dst_resource_element_configs,
            auto_initiate=auto_initiate,
            hourly_snap_replication_policy=hourly_snap_replication_policy,
            daily_snap_replication_policy=daily_snap_replication_policy,
            replicate_existing_snaps=replicate_existing_snaps)
        hc.assert_that(rs.id, hc.equal_to(
            '42949675780_FNM00150600267_0000_42949678642_FNM00152000052_0000'))

    @mock.patch_rest
    @ddt.data(
        {'max_time_out_of_sync': 30, 'name': 'modified-name',
         'hourly_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=True, is_retention_same_as_source=False,
                 is_auto_delete=False, retention_duration=3600),
         'daily_snap_replication_policy':
             replication_session.UnitySnapReplicationPolicy.to_embedded(
                 is_replicating_snaps=False),
         'src_spa_interface': 'FNM00152000052:if_7',
         'src_spb_interface': 'FNM00152000052:if_6',
         'dst_spa_interface': 'FNM00150600267:if_10',
         'dst_spb_interface': 'FNM00150600267:if_11'},
        {'max_time_out_of_sync': 30, 'name': None,
         'hourly_snap_replication_policy': None,
         'daily_snap_replication_policy': None,
         'src_spa_interface': None, 'src_spb_interface': None,
         'dst_spa_interface': None, 'dst_spb_interface': None})
    @ddt.unpack
    def test_modify(self, max_time_out_of_sync, name,
                    hourly_snap_replication_policy,
                    daily_snap_replication_policy,
                    src_spa_interface, src_spb_interface,
                    dst_spa_interface, dst_spb_interface):

        if src_spa_interface:
            src_spa_interface = remote_interface.UnityRemoteInterface(
                _id=src_spa_interface, cli=mock.t_rest())
        if src_spb_interface:
            src_spb_interface = remote_interface.UnityRemoteInterface(
                _id=src_spb_interface, cli=mock.t_rest())
        if dst_spa_interface:
            dst_spa_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spa_interface, cli=mock.t_rest())
        if dst_spb_interface:
            dst_spb_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spb_interface, cli=mock.t_rest())

        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.modify(
            max_time_out_of_sync=max_time_out_of_sync, name=name,
            hourly_snap_replication_policy=hourly_snap_replication_policy,
            daily_snap_replication_policy=daily_snap_replication_policy,
            src_spa_interface=src_spa_interface,
            src_spb_interface=src_spb_interface,
            dst_spa_interface=dst_spa_interface,
            dst_spb_interface=dst_spb_interface)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data(
        {'force_full_copy': True,
         'src_spa_interface': 'FNM00152000052:if_7',
         'src_spb_interface': 'FNM00152000052:if_6',
         'dst_spa_interface': 'FNM00150600267:if_10',
         'dst_spb_interface': 'FNM00150600267:if_11'},
        {'force_full_copy': True,
         'src_spa_interface': None,
         'src_spb_interface': None,
         'dst_spa_interface': None,
         'dst_spb_interface': None})
    @ddt.unpack
    def test_resume(self, force_full_copy,
                    src_spa_interface, src_spb_interface,
                    dst_spa_interface, dst_spb_interface):

        if src_spa_interface:
            src_spa_interface = remote_interface.UnityRemoteInterface(
                _id=src_spa_interface, cli=mock.t_rest())
        if src_spb_interface:
            src_spb_interface = remote_interface.UnityRemoteInterface(
                _id=src_spb_interface, cli=mock.t_rest())
        if dst_spa_interface:
            dst_spa_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spa_interface, cli=mock.t_rest())
        if dst_spb_interface:
            dst_spb_interface = remote_interface.UnityRemoteInterface(
                _id=dst_spb_interface, cli=mock.t_rest())
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.resume(
            force_full_copy=force_full_copy,
            src_spa_interface=src_spa_interface,
            src_spb_interface=src_spb_interface,
            dst_spa_interface=dst_spa_interface,
            dst_spb_interface=dst_spb_interface)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    def test_pause(self):
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.pause()
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    def test_sync(self):
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.sync()
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data({'sync': True, 'force': False},
              {'sync': False, 'force': None})
    @ddt.unpack
    def test_failover(self, sync, force):
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.failover(sync=sync, force=force)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))

    @mock.patch_rest
    @ddt.data({'force_full_copy': True},
              {'force_full_copy': None})
    @ddt.unpack
    def test_failback(self, force_full_copy):
        rs = replication_session.UnityReplicationSession.get(
            cli=mock.t_rest(),
            _id='42949675780_FNM00150600267_0000'
                '_42949678642_FNM00152000052_0000')
        resp = rs.failback(force_full_copy=force_full_copy)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, equal_to, raises, has_item

from storops_test.vnx.nas_mock import t_nas, patch_nas
from storops.exception import VNXGeneralNasError
from storops.vnx.resource.cifs_share import VNXCifsShare
from storops.vnx.resource.fs import VNXFileSystem
from storops.vnx.resource.mover import VNXMover

__author__ = 'Jay Xu'


class VNXCifsShareTest(unittest.TestCase):
    # todo: add test for share access commands

    @patch_nas
    def test_get_all(self):
        shares = VNXCifsShare.get(cli=t_nas())
        assert_that(len(shares), equal_to(16))
        share = next(s for s in shares if s.name == 'zhuanc_cifs_100g')
        self.verify_share_zhuanc(share)

    @staticmethod
    def verify_share_zhuanc(share):
        assert_that(share.path, equal_to(r'\zhuanc_fs_100g'))
        assert_that(share.fs_id, equal_to(211))
        assert_that(share.max_users, equal_to(10))
        assert_that(share.comment, equal_to('100g cifs share for zhuanc'))
        assert_that(share.name, equal_to('zhuanc_cifs_100g'))
        assert_that(share.mover_id, equal_to(1))
        assert_that(share.is_vdm, equal_to(False))
        assert_that(share.cifs_server_names, has_item('CIFS'))

    @patch_nas
    def test_get_by_mover(self):
        mover = self.get_mover_1()
        shares = VNXCifsShare.get(cli=t_nas(), mover=mover)
        for share in shares:
            assert_that(share.mover.get_mover_id(),
                        equal_to(mover.get_mover_id()))
            assert_that(share.mover.is_vdm, equal_to(mover.is_vdm))

    def get_mover_1(self):
        mover = VNXMover(mover_id=1, cli=t_nas())
        return mover

    @patch_nas
    def test_get_by_share_name(self):
        shares = VNXCifsShare.get(cli=t_nas(), name='zhuanc_cifs_100g')
        assert_that(len(shares), equal_to(1))
        self.verify_share_zhuanc(shares[0])

    @patch_nas
    def test_get_cifs_share(self):
        mover = self.get_mover_1()
        share = VNXCifsShare.get(name='zhuanc_cifs_100g', mover=mover,
                                 cli=t_nas())
        self.verify_share_zhuanc(share)

    @patch_nas
    def test_get_not_found(self):
        mover = self.get_mover_1()
        cifs = VNXCifsShare.get(name='not_exists', mover=mover, cli=t_nas())
        assert_that(cifs.existed, equal_to(False))

    @patch_nas
    def test_create_invalid_path(self):
        def f():
            mover = self.get_mover_1()
            VNXCifsShare.create(t_nas(), 'test_zhuanc', 'CIFS', mover,
                                path='/test_zhuanc')

        assert_that(f, raises(VNXGeneralNasError, 'Invalid path'))

    @patch_nas
    def test_create_success(self):
        fs = VNXFileSystem(name='zzz', cli=t_nas())
        mover = self.get_mover_1()
        share = VNXCifsShare.create(t_nas(), fs, 'CIFS', mover)
        assert_that(share.path, equal_to(r'\zzz'))

    @patch_nas
    def test_delete_success(self):
        share = VNXCifsShare(name='zzz', mover=self.get_mover_1(), cli=t_nas())
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_delete_not_found(self):
        def f():
            share = VNXCifsShare(name='yyy', mover=self.get_mover_1(),
                                 cli=t_nas())
            share.delete('CIFS')

        assert_that(f, raises(VNXGeneralNasError, 'No such file'))



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from persistqueue import PDict

from storops import exception
from storops.lib.tasks import PQueue
from storops.unity.enums import ThinCloneActionEnum

log = logging.getLogger(__name__)


class TCHelper(object):
    _tc_cache = {}
    _gc_candidates = {}
    _gc_background = None
    _GC_INTERVAL = 3600

    @staticmethod
    def set_up(persist_path):
        log.debug('Set up TCHelper, persist path: %s.', persist_path)
        TCHelper._tc_cache = PDict(persist_path, 'tc_cache',
                                   multithreading=True)
        TCHelper._gc_candidates = PDict(persist_path, 'gc_candidates',
                                        multithreading=True)
        TCHelper._gc_background = PQueue(persist_path,
                                         interval=TCHelper._GC_INTERVAL)
        TCHelper._gc_background.start()

    @staticmethod
    def clean_up():
        TCHelper._gc_background.stop()
        TCHelper._gc_background = None
        TCHelper._tc_cache = {}
        TCHelper._gc_candidates = {}

    @staticmethod
    def _compose_thin_clone(cli, name=None, snap=None, description=None,
                            io_limit_policy=None):
        return cli.make_body(name=name, snap=snap, description=description,
                             lunParameters=cli.make_body(
                                 ioLimitParameters=cli.make_body(
                                     ioLimitPolicy=io_limit_policy)))

    @staticmethod
    def thin_clone(cli, lun_or_snap, name, io_limit_policy=None,
                   description=None):
        from storops.unity.resource.lun import UnityLun
        from storops.unity.resource.storage_resource import \
            UnityStorageResource

        snap_to_tc = None  # None means to create a temp snap for thin clone
        if isinstance(lun_or_snap, UnityLun):
            tc_node = lun_or_snap
        else:
            # `lun_or_snap` is an UnitySnap, use it for thin clone directly.
            tc_node = lun_or_snap.lun
            snap_to_tc = lun_or_snap

        _id = lun_or_snap.get_id()
        if _id in TCHelper._tc_cache:
            tc_node = TCHelper._tc_cache[_id]
            snap_to_tc = None
            log.debug('Found %(id)s in TCHelper cache. Use it: %(cache)s as '
                      'the base to thin clone.',
                      {'id': _id, 'cache': tc_node})

        if snap_to_tc is None:
            snap_to_tc = tc_node.create_snap(name='tmp-{}'.format(name),
                                             is_auto_delete=False)
            log.debug('Temp snap used for thin clone is created: %s.',
                      snap_to_tc.get_id())

        req_body = TCHelper._compose_thin_clone(
            cli, name=name, snap=snap_to_tc, description=description,
            io_limit_policy=io_limit_policy)
        resp = cli.action(UnityStorageResource().resource_class,
                          tc_node.get_id(), 'createLunThinClone', **req_body)
        if snap_to_tc != lun_or_snap:
            # Delete the temp snap
            snap_to_tc.delete()
        resp.raise_if_err()
        return UnityLun(cli=cli, _id=resp.resource_id)

    @staticmethod
    def _gc_resources(lun_or_snap):
        lun_or_snap_id = lun_or_snap.get_id()
        if lun_or_snap_id in TCHelper._tc_cache:
            base_lun = TCHelper._tc_cache[lun_or_snap_id]
            log.debug('Found %(id)s in TCHelper cache. Base lun: %(base)s.',
                      {'id': lun_or_snap_id, 'base': base_lun.get_id()})
            TCHelper._gc_base_lun(base_lun)

    @staticmethod
    def _gc_base_lun(base_lun):
        TCHelper._gc_candidates[base_lun.get_id()] = base_lun.get_id()
        log.debug('Put base lun: %s to gc candidates list.',
                  base_lun.get_id())
        if TCHelper._gc_background is not None:
            TCHelper._gc_background.put(TCHelper._delete_base_lun,
                                        base_lun=base_lun)
            log.debug('Put base lun: %s to background gc process.',
                      base_lun.get_id())
        else:
            log.warning('LUN will not be deleted because background GC thread'
                        'is not running. Invoke TCHelper.set_up first.')

    @staticmethod
    def _delete_thin_clone(thin_clone):
        try:
            TCHelper._delete_base_lun(thin_clone.family_base_lun)
        except exception.UnityBaseHasThinCloneError:
            log.info('Failed to delete base lun %s due to it still has other '
                     'thin clones. The background gc will delete it later.',
                     thin_clone.family_base_lun.get_id())
        except exception.UnityTCSnapUnderDestroyError:
            log.info('Failed to delete base lun %s due to the underlying '
                     'snapshot for thin clone still under destroying. '
                     'The background gc will delete it later.',
                     thin_clone.family_base_lun.get_id())

    @classmethod
    def _delete_base_lun(cls, base_lun):
        if base_lun.get_id() in TCHelper._gc_candidates:
            if base_lun.family_clone_count > 0:
                # Raising exception will let the job re-enter the background gc
                # queue
                raise exception.UnityBaseHasThinCloneError()

            base_lun.delete()
            del TCHelper._gc_candidates[base_lun.get_id()]
            log.debug('Base lun %s deleted. And removed from gc candidates.',
                      base_lun.get_id())

    @staticmethod
    def _update_cache(lun_or_snap, action_enum, *args):
        lun_or_snap_id = lun_or_snap.get_id()
        if action_enum in (ThinCloneActionEnum.DD_COPY,):
            TCHelper._gc_resources(lun_or_snap)
            log.debug('Garbage collected for base lun of %(id)s, triggered by '
                      '%(action)s action.',
                      {'id': lun_or_snap_id, 'action': action_enum.name})
            TCHelper._tc_cache[lun_or_snap_id] = args[0]
            log.debug('Cache updated for base lun of %(id)s to %(new)s.',
                      {'id': lun_or_snap_id, 'new': args[0]})
        elif action_enum in (ThinCloneActionEnum.LUN_ATTACH,):
            TCHelper._gc_resources(lun_or_snap)
            log.debug('Garbage collected for base lun of %(id)s, triggered by '
                      '%(action)s action.',
                      {'id': lun_or_snap_id, 'action': action_enum.name})
            if lun_or_snap_id in TCHelper._tc_cache:
                del TCHelper._tc_cache[lun_or_snap_id]
                log.debug('Removed base lun of %s from cache.', lun_or_snap_id)
        elif action_enum in (ThinCloneActionEnum.BASE_LUN_DELETE,):
            # `lun_or_snap` is a base lun here.
            TCHelper._gc_base_lun(lun_or_snap)
            log.debug('Garbage collected for base lun of %(id)s, triggered by '
                      '%(action)s action.',
                      {'id': lun_or_snap_id, 'action': action_enum.name})

    @staticmethod
    def notify(lun_or_snap, action_enum, *args):
        if action_enum in (ThinCloneActionEnum.TC_DELETE,):
            log.debug('Try to delete base lun of %s actively.',
                      lun_or_snap.get_id())
            # Delete the base lun of thin-clone actively.
            TCHelper._delete_thin_clone(lun_or_snap)

        TCHelper._update_cache(lun_or_snap, action_enum, *args)



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Ryan Liang'

log = logging.getLogger(__name__)


class UnityReplicationInterface(UnityResource):
    @classmethod
    def create(cls, cli, sp, ip_port, ip_address, netmask=None,
               v6_prefix_length=None, gateway=None, vlan_id=None):
        """
        Creates a replication interface.

        :param cls: this class.
        :param cli: the rest cli.
        :param sp: `UnityStorageProcessor` object. Storage processor on which
            the replication interface is running.
        :param ip_port: `UnityIpPort` object. Physical port or link aggregation
            on the storage processor on which the interface is running.
        :param ip_address: IP address of the replication interface.
        :param netmask: IPv4 netmask for the replication interface, if it uses
            an IPv4 address.
        :param v6_prefix_length: IPv6 prefix length for the interface, if it
            uses an IPv6 address.
        :param gateway: IPv4 or IPv6 gateway address for the replication
            interface.
        :param vlan_id: VLAN identifier for the interface.
        :return: the newly create replication interface.
        """
        req_body = cli.make_body(sp=sp, ipPort=ip_port,
                                 ipAddress=ip_address, netmask=netmask,
                                 v6PrefixLength=v6_prefix_length,
                                 gateway=gateway, vlanId=vlan_id)

        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        return cls.get(cli, resp.resource_id)

    def modify(self, sp=None, ip_port=None, ip_address=None, netmask=None,
               v6_prefix_length=None, gateway=None, vlan_id=None):
        """
        Modifies a replication interface.

        :param sp: same as the one in `create` method.
        :param ip_port: same as the one in `create` method.
        :param ip_address: same as the one in `create` method.
        :param netmask: same as the one in `create` method.
        :param v6_prefix_length: same as the one in `create` method.
        :param gateway: same as the one in `create` method.
        :param vlan_id: same as the one in `create` method.
        """
        req_body = self._cli.make_body(sp=sp, ipPort=ip_port,
                                       ipAddress=ip_address, netmask=netmask,
                                       v6PrefixLength=v6_prefix_length,
                                       gateway=gateway, vlanId=vlan_id)

        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp


class UnityReplicationInterfaceList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityReplicationInterface



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of

from storops.unity.enums import SNMPPrivacyProtocolEnum, SNMPAuthProtocolEnum
from storops.unity.resource.alert_config_snmp import UnityAlertConfigSNMPTarget
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Dong Ding'


class UnityAlertConfigSNMPTargetTest(TestCase):
    @patch_rest
    def test_create_alert_snmp_v3_config(self):
        alert_snmp_config = UnityAlertConfigSNMPTarget.create(
            cli=t_rest(),
            target_address='10.10.10.111',
            username='test_username',
            auth_protocol=SNMPAuthProtocolEnum.MD5,
            priv_protocol=SNMPPrivacyProtocolEnum.AES,
            auth_password='auth_password_test')
        assert_that(alert_snmp_config, instance_of(
            UnityAlertConfigSNMPTarget))
        assert_that(alert_snmp_config.get_id(), equal_to(
            'snmp_target_8'))
        assert_that(alert_snmp_config.address, equal_to(
            '10.10.10.111'))

    @patch_rest
    def test_create_alert_snmp_v2_config(self):
        alert_snmp_config = UnityAlertConfigSNMPTarget.create(
            cli=t_rest(), target_address='10.10.10.222',
            community='test_community')
        assert_that(alert_snmp_config, instance_of(
            UnityAlertConfigSNMPTarget))
        assert_that(alert_snmp_config.get_id(), equal_to(
            'snmp_target_9'))

    @patch_rest
    def test_modify_alert_snmp_config(self):
        alert_snmp_config = UnityAlertConfigSNMPTarget.create(
            cli=t_rest(), target_address='10.10.10.111',
            username='test_username_2')
        resp = alert_snmp_config.modify(username='test_username_modified_2')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_alert_snmp_config(self):
        alert_snmp_config = UnityAlertConfigSNMPTarget.create(
            cli=t_rest(), target_address='10.10.10.111',
            username='test_username_2')
        resp = alert_snmp_config.delete()
        assert_that(resp.body, equal_to({}))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import equal_to, assert_that, only_contains, instance_of, \
    raises, none
from storops.unity.resource.snap import UnitySnap

from storops.exception import UnityException, UnityNfsShareNameExistedError, \
    UnityHostNotFoundException, UnitySnapNameInUseError
from storops.unity.enums import NFSTypeEnum, NFSShareRoleEnum, \
    NFSShareDefaultAccessEnum, NFSShareSecurityEnum
from storops.unity.resource.filesystem import UnityFileSystem
from storops.unity.resource.host import UnityHostList, UnityHost
from storops.unity.resource.nfs_share import UnityNfsShare, \
    UnityNfsShareList, UnityNfsHostConfig
from storops_test.unity.rest_mock import patch_rest, t_rest

__author__ = 'Cedric Zhuang'


def host(_id):
    return UnityHost(_id=_id, cli=t_rest())


class UnityNfsShareTest(TestCase):
    @patch_rest
    def test_properties(self):
        nfs = UnityNfsShare('NFSShare_1', cli=t_rest())
        assert_that(nfs.id, equal_to('NFSShare_1'))
        assert_that(nfs.type, equal_to(NFSTypeEnum.NFS_SHARE))
        assert_that(nfs.role, equal_to(NFSShareRoleEnum.PRODUCTION))
        assert_that(nfs.default_access,
                    equal_to(NFSShareDefaultAccessEnum.ROOT))
        assert_that(nfs.min_security, equal_to(NFSShareSecurityEnum.SYS))
        assert_that(nfs.name, equal_to('esa_nfs1'))
        assert_that(nfs.path, equal_to(r'/'))
        assert_that(nfs.export_paths,
                    only_contains(r'10.244.220.120:/esa_nfs1'))
        assert_that(nfs.description, equal_to('bcd'))
        assert_that(str(nfs.creation_time),
                    equal_to('2016-03-02 02:39:22.856000+00:00'))
        assert_that(str(nfs.modification_time),
                    equal_to('2016-03-02 02:39:22.856000+00:00'))
        assert_that(nfs.filesystem.get_id(), equal_to('fs_1'))
        assert_that(nfs.filesystem, instance_of(UnityFileSystem))
        assert_that(nfs.tenant, equal_to(None))

    @patch_rest
    def test_tenant_properties(self):
        nfs = UnityNfsShare('NFSShare_32', cli=t_rest())
        assert_that(nfs.tenant.id, equal_to('tenant_1'))

    @patch_rest
    def test_host_access_properties(self):
        nfs = UnityNfsShare('NFSShare_5', cli=t_rest())
        assert_that(nfs.read_write_hosts, only_contains(host('Host_7')))
        assert_that(nfs.read_only_hosts, none())
        assert_that(nfs.root_access_hosts, none())
        assert_that(nfs.no_access_hosts, none())

    @patch_rest
    def test_get_all(self):
        nfs_list = UnityNfsShareList(cli=t_rest())
        assert_that(len(nfs_list), equal_to(2))

    @patch_rest
    def test_create_nfs_share_fs_not_support(self):
        def f():
            UnityNfsShare.create(t_rest(), 'ns1', 'fs_8')

        assert_that(f, raises(UnityException, 'not support NFS'))

    @patch_rest
    def test_create_nfs_share_success(self):
        share = UnityNfsShare.create(
            t_rest(), 'ns1', 'fs_9',
            share_access=NFSShareDefaultAccessEnum.READ_WRITE)
        assert_that(share.name, equal_to('ns1'))
        assert_that(share.id, equal_to('NFSShare_4'))

    @patch_rest
    def test_create_nfs_share_ro_root(self):
        share = UnityNfsShare.create(
            t_rest(), 'ns1', 'fs_9',
            share_access=NFSShareDefaultAccessEnum.RO_ROOT)
        assert_that(share.name, equal_to('ns1'))
        assert_that(share.id, equal_to('NFSShare_4'))

    @patch_rest
    def test_create_nfs_share_name_exists(self):
        def f():
            UnityNfsShare.create(
                t_rest(), 'ns1', 'fs_9',
                share_access=NFSShareDefaultAccessEnum.ROOT)

        assert_that(f, raises(UnityNfsShareNameExistedError, 'already exists'))

    @patch_rest
    def test_delete_nfs_share_success(self):
        share = UnityNfsShare(_id='NFSShare_4', cli=t_rest())
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_nfs_share_snap_existed(self):
        def f():
            share = UnityNfsShare(_id='NFSShare_31', cli=t_rest())
            share.create_snap('share_snap')

        assert_that(f, raises(UnitySnapNameInUseError, 'in use'))

    @patch_rest
    def test_delete_nfs_share_async(self):
        share = UnityNfsShare(_id='NFSShare_6', cli=t_rest())
        resp = share.delete(async_mode=True)
        job = resp.job
        assert_that(job.existed, equal_to(True))
        assert_that(str(job.est_remain_time), equal_to('0:00:01'))

    @patch_rest
    def test_delete_snap_based_share(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_11')
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_modify_read_write_hosts(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_5')
        resp = share.modify(read_write_hosts=['Host_7'])
        assert_that(resp.is_ok(), equal_to(True))
        share.update()
        assert_that(share.read_write_hosts, instance_of(UnityHostList))
        assert_that(share.read_write_hosts[0].get_id(), equal_to('Host_7'))

    @patch_rest
    def test_modify_multiple_hosts(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_5')
        host7 = UnityHost(_id='Host_7', cli=t_rest())
        resp = share.modify(read_only_hosts=host7,
                            read_write_hosts=['Host_1', 'Host_2'])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_ip_access_force_create(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_5')
        h1 = UnityHost(_id='Host_1', cli=t_rest())
        resp = share.allow_read_write_access([h1, '1.1.1.2'],
                                             force_create_host=True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_ip_access_existed(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_5')
        h1 = UnityHost(_id='Host_1', cli=t_rest())
        resp = share.allow_read_only_access([h1, '1.1.1.1'],
                                            force_create_host=True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_share_in_tenant_allow_access_to_ip(self):
        share = UnityNfsShare(_id='NFSShare_32', cli=t_rest())
        resp = share.allow_read_only_access('192.168.112.23')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_access_success(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_7')
        resp = share.delete_access(['Host_1', 'Host_14', 'Host_15'])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_access_ip_not_found(self):
        def f():
            share = UnityNfsShare(cli=t_rest(), _id='NFSShare_7')
            share.delete_access(['9.9.9.9'])

        assert_that(f, raises(UnityHostNotFoundException, 'not found'))

    @patch_rest
    def test_delete_access_host_not_found(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_4')
        resp = share.delete_access(['Host_99'])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_clear_access(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_7')
        resp = share.clear_access()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_read_only_snap_share(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_30')
        resp = share.allow_read_only_access('1.1.1.1', True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_read_write_snap_share(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_30')
        resp = share.allow_read_write_access('2.2.2.2', True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_access_snap_share(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_31')
        resp = share.delete_access(['1.1.1.1', '1.1.1.3'])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_snap_based_share_properties(self):
        share = UnityNfsShare.get(t_rest(), _id='NFSShare_2')
        assert_that(share.type, equal_to(NFSTypeEnum.NFS_SNAPSHOT))
        assert_that(share.filesystem, instance_of(UnityFileSystem))
        assert_that(share.snap, instance_of(UnitySnap))


class UnityNfsHostConfigTest(TestCase):
    def test_add_ro(self):
        config = UnityNfsHostConfig(
            no_access=[host('Host_1'), host('Host_9')])
        config.allow_ro(host('Host_1'), host('Host_11'))
        config.allow_rw(host('Host_9'))

        assert_that(config.root,
                    only_contains(host('Host_1'),
                                  host('Host_9'),
                                  host('Host_11')))
        assert_that(config.ro,
                    only_contains(host('Host_1'),
                                  host('Host_11')))
        assert_that(config.rw, only_contains(host('Host_9')))
        assert_that(len(config.no_access), equal_to(0))

    def test_add_same_twice(self):
        config = UnityNfsHostConfig(no_access=[host('Host_9')])
        config.allow_rw(host('Host_9'), host('Host_9'))
        config.allow_rw(host('Host_9'))
        assert_that(len(config.rw), equal_to(1))
        assert_that(config.rw, only_contains(host('Host_9')))
        assert_that(config.no_access, equal_to([]))
        assert_that(config.ro, none())
        assert_that(config.root, only_contains(host('Host_9')))

    def test_clear_access_all(self):
        config = UnityNfsHostConfig(no_access=[host('Host_9')])
        config.allow_rw(host('Host_1'))
        config.clear_all()
        assert_that(len(config.rw), equal_to(0))
        assert_that(len(config.ro), equal_to(0))
        assert_that(len(config.no_access), equal_to(0))
        assert_that(len(config.root), equal_to(0))

    def test_clear_access_with_white_list(self):
        config = UnityNfsHostConfig(no_access=[host('Host_9')])
        config.allow_rw(host('Host_1'), host('Host_2'))
        config.allow_root(host('Host_1'), host('Host_2'), host('Host_3'))
        config.clear_all(host('Host_1'), host('Host_4'), host('Host_9'))
        assert_that(len(config.rw), equal_to(0))
        assert_that(len(config.ro), equal_to(0))
        assert_that(len(config.no_access), equal_to(1))
        assert_that(len(config.root), equal_to(1))



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import threading
from collections import OrderedDict
from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, raises
from mock import patch

from storops.exception import JobStateError, JobTimeoutException
from storops.lib import job_helper
from storops.unity.resource.job import UnityJob
from storops_test.unity.rest_mock import t_rest, patch_rest


class MockJob(object):
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)


class TestJobHelper(TestCase):
    def setUp(self):
        self.jh = job_helper.JobHelper(cli=t_rest())

    def tearDown(self):
        self.jh.jobs = OrderedDict()

    def test_add_job(self):
        job_id = 'N-101'
        job = MockJob(id=job_id)
        self.jh.add_job(job)
        assert_that(self.jh.jobs[job_id].id, equal_to(job_id))

    def test_add_existed_job(self):
        job_id = 'N-102'
        job = MockJob(id=job_id)
        self.jh.add_job(job)
        self.jh.add_job(job)
        assert_that(self.jh.jobs[job_id].id, equal_to(job_id))

    def test_remove_job(self):
        job_id = 'N-103'
        job = MockJob(id=job_id)
        self.jh.jobs[job_id] = job
        self.jh.remove_job(job)
        ret = job_id in self.jh.jobs
        assert_that(ret, equal_to(False))

    def test_remove_nonexisted_job(self):
        job_id = 'N-104'
        job = MockJob(id=job_id)
        self.jh.remove_job(job)
        ret = job_id in self.jh.jobs
        assert_that(ret, equal_to(False))

    def test_get_job(self):
        job_id = 'N-105'
        job = MockJob(id=job_id)
        self.jh.jobs[job_id] = job
        assert_that(self.jh.get_job(job).id, equal_to(job_id))

    def test_get_nonexisted_job(self):
        job_id = 'N-106'
        job = MockJob(id=job_id)
        assert_that(self.jh.get_job(job), equal_to(None))

    @patch_rest
    def test_wait_job(self):
        job_id = 'N-3078'
        job = UnityJob(_id=job_id, cli=t_rest())
        new_job = self.jh.wait_job(job, 10, 3)
        assert_that(new_job.id, equal_to(job_id))
        assert_that(new_job.state.index, equal_to(4))
        ret = job_id in self.jh.jobs
        assert_that(ret, equal_to(False))

    @patch_rest
    def test_wait_job_state_error(self):
        job_id = 'N-3079'

        def f():
            job = UnityJob(_id=job_id, cli=t_rest())
            self.jh.wait_job(job, 10, 3)

        assert_that(f, raises(JobStateError))
        ret = job_id in self.jh.jobs
        assert_that(ret, equal_to(False))

    @patch_rest
    def test_wait_job_state_timeout(self):
        job_id = 'N-3080'

        def f():
            job = UnityJob(_id=job_id, cli=t_rest())
            self.jh.wait_job(job, 3, 1)

        assert_that(f, raises(JobTimeoutException))
        ret = job_id in self.jh.jobs
        assert_that(ret, equal_to(False))

    @patch_rest
    def test__query_jobs(self):
        job_1 = MockJob(id='N-3074')
        job_2 = MockJob(id='N-3075')
        job_3 = MockJob(id='N-3076')
        self.jh.add_job(job_1)
        self.jh.add_job(job_2)
        self.jh.add_job(job_3)
        self.jh._query_jobs()
        new_job_1 = self.jh.get_job(job_1)
        new_job_2 = self.jh.get_job(job_2)
        new_job_3 = self.jh.get_job(job_3)
        assert_that(new_job_1.id, equal_to(job_1.id))
        assert_that(new_job_1.state.index, equal_to(4))
        assert_that(new_job_2.id, equal_to(job_2.id))
        assert_that(new_job_2.state.index, equal_to(4))
        assert_that(new_job_3.id, equal_to(job_3.id))
        assert_that(new_job_3.state.index, equal_to(4))

    def test__query_jobs_no_job(self):
        self.jh.jobs = OrderedDict()
        self.jh._query_jobs()
        assert_that(self.jh.jobs, equal_to(OrderedDict()))

    def test_start(self):
        self.jh.start()
        assert_that(self.jh.started, equal_to(True))
        assert_that(self.jh.t, instance_of(threading.Thread))
        assert_that(self.jh.t.daemon, equal_to(True))


class TestGetJobHelper(TestCase):
    @patch('threading.Thread')
    def test_get_job_helper(self, mocked_thread):
        jh = job_helper.get_job_helper(cli=t_rest())
        assert_that(jh.jobs, equal_to(OrderedDict()))
        assert_that(jh.started, equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import unittest

from hamcrest import assert_that, not_none, equal_to, instance_of

import storops

__author__ = 'Cedric Zhuang'

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
    level=logging.INFO)


class StoropsTest(unittest.TestCase):
    def test_vnx_availability(self):
        vnx = storops.VNXSystem('10.244.211.30', heartbeat_interval=0)
        assert_that(vnx, not_none())

    def test_unity_availability(self):
        unity = storops.UnitySystem('1.1.1.1', 'admin', 'password')
        assert_that(unity, not_none())

    def test_vnx_enum_availability(self):
        spa = storops.VNXSPEnum.SP_A
        assert_that(spa, not_none())

    def test_unity_enum_availability(self):
        raid5 = storops.RaidTypeEnum.RAID5
        assert_that(raid5, not_none())

    def test_enable_log(self):
        storops.enable_log()
        log = logging.getLogger('storops')
        assert_that(len(log.handlers), equal_to(1))
        assert_that(log.handlers[0], instance_of(logging.StreamHandler))
        assert_that(log.getEffectiveLevel(), equal_to(logging.DEBUG))

    def test_enable_log_called_twice(self):
        storops.enable_log()
        storops.enable_log(logging.INFO)
        log = logging.getLogger('storops')
        assert_that(len(log.handlers), equal_to(1))
        assert_that(log.handlers[0], instance_of(logging.StreamHandler))
        assert_that(log.getEffectiveLevel(), equal_to(logging.INFO))

    def test_disable_log(self):
        storops.enable_log()
        log = storops.disable_log()
        assert_that(len(log.handlers), equal_to(0))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from storops_comptest import t_vnx
from hamcrest import assert_that, equal_to

__author__ = 'Cedric Zhuang'

vnx = t_vnx()


def _create_cg_with_member(vnx_gf):
    """ Helper function to create a cg with a lun

    :param vnx_gf: the vnx general test fixture
    :return: created cg
    """
    lun_name = vnx_gf.add_lun_name()
    lun = vnx_gf.pool.create_lun(lun_name)

    cg_name = vnx_gf.add_cg_name()
    cg = vnx_gf.vnx.create_cg(cg_name)
    cg.add_member(lun)
    return cg


def test_cg_create_success(vnx_gf):
    assert_that(vnx_gf.cg.existed, equal_to(True))


def test_add_delete_cg_member(vnx_gf):
    lun_name = vnx_gf.add_lun_name()
    lun = vnx_gf.pool.create_lun(lun_name)

    # add member
    vnx_gf.cg.add_member(lun)
    vnx_gf.cg.update()
    assert_that(vnx_gf.cg.has_member(lun), equal_to(True))

    # delete member
    vnx_gf.cg.delete_member(lun)
    vnx_gf.cg.update()
    assert_that(vnx_gf.cg.has_member(lun), equal_to(False))


def test_update_cg_member(vnx_gf):
    lun_name = vnx_gf.add_lun_name()
    lun = vnx_gf.pool.create_lun(lun_name)

    # update member
    cg = _create_cg_with_member(vnx_gf)
    assert_that(cg.has_member(lun), equal_to(False))

    cg.replace_member(lun)
    cg.update()
    assert_that(cg.has_member(lun), equal_to(True))


@pytest.mark.skipif(not vnx.is_snap_enabled(),
                    reason='snap feature not available')
def test_create_delete_cg_snapshot(vnx_gf):
    cg = _create_cg_with_member(vnx_gf)

    # create snap
    name = vnx_gf.add_snap_name()
    snap = cg.create_snap(name)
    assert_that(snap.name, equal_to(name))
    assert_that(snap.existed, equal_to(True))

    # delete snap
    snap.delete()
    snap.update()
    assert_that(snap.existed, equal_to(False))


def test_delete_cg_success(vnx_gf):
    name = vnx_gf.add_cg_name()
    cg = vnx_gf.vnx.create_cg(name)
    assert_that(cg.existed, equal_to(True))

    cg.delete()
    cg.update()
    assert_that(cg.existed, equal_to(False))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import random
from threading import Lock

from retryz import retry

import storops.vnx.resource.host
import storops.vnx.resource.lun
from storops import exception as ex
from storops.lib.common import instance_cache
from storops.vnx.enums import VNXPortType
from storops.vnx.resource import VNXCliResource, VNXCliResourceList
from storops.vnx.resource.port import VNXHbaPort

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXStorageGroup(VNXCliResource):
    def __init__(self, name=None, cli=None, shuffle_hlu=True,
                 system_lun_list=None):
        super(VNXStorageGroup, self).__init__()
        self._cli = cli
        self._name = name
        self._system_lun_list = system_lun_list

        self._uid = ''
        self._hba_port_list = []
        self._conn = None
        self._hlu_lock = Lock()
        self._alu_hlu_cache = {}

        self.shuffle_hlu = shuffle_hlu

    def _get_raw_resource(self):
        return self._cli.get_sg(name=self._name, poll=self.poll,
                                engineering=True)

    @classmethod
    def get(cls, cli, name=None, system_lun_list=None):
        if name is None:
            ret = VNXStorageGroupList(cli, system_lun_list=system_lun_list)
        else:
            ret = VNXStorageGroup(name, cli, system_lun_list=system_lun_list)
        return ret

    @classmethod
    def create(cls, name, cli):
        out = cli.create_sg(name)
        msg = 'failed to create storage group "{}".'.format(name)
        ex.raise_if_err(out, msg, default=ex.VNXCreateStorageGroupError)
        return VNXStorageGroup(name, cli)

    def delete(self, disconnect_host=False):
        if disconnect_host:
            for hba in self.hba_sp_pairs:
                self.disconnect_host(hba.host_name)
        out = self._cli.delete_sg(self._get_name(), poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXDeleteStorageGroupError)

    def has_hlu(self, hlu):
        return hlu in self.used_hlu_numbers

    def has_alu(self, lun):
        try:
            alu = storops.vnx.resource.lun.VNXLun.get_id(lun)
        except ValueError:
            # lun not found, id is None
            alu = None
        return alu is not None and alu in self.used_alu_numbers

    def get_hlu(self, lun):
        alu = storops.vnx.resource.lun.VNXLun.get_id(lun)
        return self.get_alu_hlu_map().get(alu, None)

    def is_valid(self):
        return len(self.name) > 0 and len(self.uid) > 0

    @property
    def hosts(self):
        host_names = [pair.host_name for pair in self.hba_sp_pairs]
        clz = storops.vnx.resource.host.VNXHostList
        return clz(cli=self._cli, names=host_names)

    @property
    @instance_cache
    def lun_list(self):
        lun_ids = self.get_alu_hlu_map().keys()
        if self._system_lun_list is not None:
            ret = self._system_lun_list.shadow_copy(lun_ids=lun_ids)
        else:
            clz = storops.vnx.resource.lun.VNXLunList
            ret = clz(cli=self._cli, lun_ids=lun_ids)
        ret.poll = self.poll
        return ret

    @property
    def hba_port_list(self):
        if not self._hba_port_list:
            self.hba_port_list = self.hba_sp_pairs
        return self._hba_port_list

    @hba_port_list.setter
    def hba_port_list(self, hba_sp_pairs):
        def _process_cli_output(value):
            port = VNXHbaPort.from_storage_group_hba(value)
            hba = value.uid
            self._hba_port_list.append((hba, port))

        if hba_sp_pairs is not None:
            for item in hba_sp_pairs:
                _process_cli_output(item)

    @staticmethod
    def _port_filter(ports, sp=None, port_id=None, vport_id=None,
                     port_type=None):
        def _filter(port):
            ret = True
            if sp is not None:
                ret &= port.sp == sp
            if port_id is not None:
                ret &= port.port_id == port_id
            if vport_id is not None:
                ret &= port.vport_id == vport_id
            if port_type is not None:
                ret &= port.type == port_type
            return ret

        return tuple(filter(_filter, ports))

    def _get_hba_ports(self, sp=None, port_id=None, vport_id=None,
                       port_type=None):
        ports = set(map(lambda x: x[1], self.hba_port_list))
        return self._port_filter(ports, sp, port_id, vport_id, port_type)

    def get_ports(self, initiator_uid=None, sp=None, port_id=None,
                  vport_id=None, port_type=None):
        if initiator_uid is not None:
            ret = []
            for hba, port in self.hba_port_list:
                if hba == initiator_uid:
                    ret.append(port)
            ret = tuple(set(ret))
        else:
            ret = self._get_hba_ports(sp, port_id, vport_id, port_type)
        return ret

    def get_fc_ports(self, initiator_uid=None, sp=None, port_id=None,
                     vport_id=None):
        return self.get_ports(initiator_uid=initiator_uid,
                              sp=sp,
                              port_id=port_id,
                              vport_id=vport_id,
                              port_type=VNXPortType.FC)

    def get_iscsi_ports(self, initiator_uid=None, sp=None, port_id=None,
                        vport_id=None):
        return self.get_ports(initiator_uid=initiator_uid,
                              sp=sp,
                              port_id=port_id,
                              vport_id=vport_id,
                              port_type=VNXPortType.ISCSI)

    @property
    def ports(self):
        return self.get_ports()

    @property
    def iscsi_ports(self):
        return self.get_ports(port_type=VNXPortType.ISCSI)

    @property
    def fc_ports(self):
        return self.get_ports(port_type=VNXPortType.FC)

    @property
    def initiator_uid_list(self):
        if self.hba_sp_pairs is not None:
            ret = tuple(set(map(lambda x: x.uid, self.hba_sp_pairs)))
        else:
            ret = tuple()
        return ret

    def get_initiator_uids(self, port_type=None):
        ret = []
        for hba, port in self.hba_port_list:
            if port_type is not None:
                if port.type == port_type:
                    ret.append(hba)
            else:
                ret.append(hba)
        return tuple(set(ret))

    _hlu_full = None
    _max_hlu = 255

    @classmethod
    def get_max_luns_per_sg(cls):
        return cls._max_hlu

    @classmethod
    def set_max_luns_per_sg(cls, value):
        log.info('Update max LUNs per storage group to: {}'
                 .format(value))
        cls._max_hlu = value
        cls._hlu_full = None

    @classmethod
    def _hlu_full_set(cls):
        if cls._hlu_full is None:
            cls._hlu_full = set(range(1, cls.get_max_luns_per_sg() + 1))
        return set(cls._hlu_full)

    def get_alu_hlu_map(self):
        if self.alu_hlu_map is None:
            self._parsed_resource['alu_hlu_map'] = {}
        return self._alu_hlu_cache

    @property
    def used_hlu_numbers(self):
        return self.get_alu_hlu_map().values()

    @property
    def used_alu_numbers(self):
        return self.get_alu_hlu_map().keys()

    def _get_hlu_to_add(self, alu):
        ret = None
        with self._hlu_lock:
            remain = self._hlu_full_set() - set(self.used_hlu_numbers)
            if len(remain) == 0:
                raise ex.VNXNoHluAvailableError(
                    'no hlu number available for attach.')
            if self.shuffle_hlu:
                ret = random.sample(remain, 1)[0]
            else:
                ret = remain.pop()
            self.get_alu_hlu_map()[alu] = ret
        return ret

    def _add_hlu(self, alu, hlu):
        with self._hlu_lock:
            self.get_alu_hlu_map()[alu] = hlu
        return hlu

    def _delete_alu(self, alu):
        ret = None
        with self._hlu_lock:
            if self.has_alu(alu):
                ret = self.get_alu_hlu_map().pop(alu)
        return ret

    def update(self, data=None):
        ret = super(VNXStorageGroup, self).update(data)

        # only update the _alu_hlu_cache incrementally.
        if self.alu_hlu_map:
            self._alu_hlu_cache.update(self.alu_hlu_map)
        return ret

    def attach_alu(self, lun, retry_limit=None, hlu=None):
        def _update():
            self.update()

        @retry(on_error=ex.VNXHluNumberInUseError, on_retry=_update,
               limit=retry_limit)
        def _do(alu_id, hlu_id=None):
            if hlu_id is None:
                hlu_id = self._get_hlu_to_add(alu_id)
            elif hlu_id in self.used_hlu_numbers:
                raise ex.VNXHluAlreadyUsedError('hlu id is already used.')
            else:
                hlu_id = self._add_hlu(alu_id, hlu_id)
            out = self._cli.sg_add_hlu(self._get_name(), hlu_id, alu_id,
                                       poll=self.poll)
            try:
                ex.raise_if_err(out, default=ex.VNXAttachAluError)
            except ex.VNXAluAlreadyAttachedError:
                # alu no in the alu-hlu map cache but attach failed with
                # already attached, that means the cache is out dated
                log.info("ALU(alu=%s) is already attached, "
                         "updating the cache.", alu)
                self.update()
                raise
            except ex.VNXAttachAluError:
                # other attach error, remove hlu id from the cache
                log.info("ALU(alu=%s) failed to attach, "
                         "deleting from the cache.", alu)
                self._delete_alu(alu_id)
                raise

            # update the cache with alu/hlu again, in case it was flush out
            # by an "update" call.
            self._add_hlu(alu_id, hlu_id)
            return hlu_id

        lun_clz = storops.vnx.resource.lun.VNXLun
        alu = lun_clz.get_id(lun)
        if self.has_alu(alu):
            # found alu in the alu-hlu map cache, meaning already attached
            log.info("ALU(alu=%s) is already attached, found in the cache.",
                     alu)
            raise ex.VNXAluAlreadyAttachedError()
        else:
            ret = _do(alu, hlu)
        return ret

    def detach_alu(self, lun):
        alu = storops.vnx.resource.lun.VNXLun.get_id(lun)
        hlu = self.get_hlu(lun)
        if hlu is None:
            raise ex.VNXDetachAluNotFoundError(
                'specified lun {} is not attached.'.format(alu))
        out = self._cli.sg_delete_hlu(self._get_name(), hlu, poll=self.poll)
        msg = 'failed to detach hlu {}/alu {}.'.format(hlu, alu)
        ex.raise_if_err(out, msg, default=ex.VNXStorageGroupError, )
        self._delete_alu(alu)

    def connect_host(self, host):
        out = self._cli.sg_connect_host(self._get_name(), host, poll=self.poll)
        msg = 'failed to connect host {}.'.format(host)
        ex.raise_if_err(out, msg, default=ex.VNXStorageGroupError)

    def disconnect_host(self, host):
        out = self._cli.sg_disconnect_host(self._get_name(), host,
                                           poll=self.poll)
        msg = 'failed to disconnect host {}.'.format(host)
        ex.raise_if_err(out, msg, default=ex.VNXStorageGroupError)

    def connect_hba(self, port, hba_uid, host_name, host_ip=None):
        return self.set_path(port, hba_uid, host_name, host_ip)

    @property
    def uid(self):
        return self.wwn

    def set_path(self, port, hba_uid, host_name, host_ip=None):
        if hasattr(port, 'sp'):
            sp = port.sp
        else:
            raise ValueError('sp is not available from {}.'.format(port))

        if hasattr(port, 'port_id'):
            port_id = port.port_id
        else:
            raise ValueError('port id is not available from {}.'.format(port))

        vport_id = None
        if hasattr(port, 'virtual_port_id'):
            # FCoE do not need vport_id
            if hasattr(port, 'type'):
                port_type = port.type
                if port_type == VNXPortType.ISCSI:
                    vport_id = port.virtual_port_id

        out = self._cli.set_path(self._get_name(), hba_uid, sp, port_id,
                                 host_ip, host_name, vport_id=vport_id)
        ex.raise_if_err(out)

    def set_system_lun_list(self, system_lun_list):
        self._system_lun_list = system_lun_list


class VNXStorageGroupList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXStorageGroup

    def __init__(self, cli=None, engineering=False, system_lun_list=None,
                 attached_lun=None):
        super(VNXStorageGroupList, self).__init__(cli)
        self._sg_map = {}
        self._engineering = engineering
        self._system_lun_list = system_lun_list

        self._attached_lun = None
        self._set_filter(attached_lun)

    def _set_filter(self, attached_lun=None):
        self._attached_lun = attached_lun

    def _filter(self, sg):
        if self._attached_lun is not None:
            ret = self._attached_lun in sg.used_alu_numbers
        else:
            ret = True
        return ret

    def add_sg(self, sg):
        self._sg_map[sg.name] = sg

    def detach_alu(self, lun):
        for sg in self:
            if sg.has_alu(lun):
                sg.detach_alu(lun)

    def _get_raw_resource(self):
        return self._cli.get_sg(poll=self.poll, engineering=self._engineering)

    @property
    @instance_cache
    def _name_sg_map(self):
        return {sg.name: sg for sg in self}

    def _get_resource_instance(self):
        ret = super(VNXStorageGroupList, self)._get_resource_instance()
        ret.set_system_lun_list(self._system_lun_list)
        return ret

    def get(self, name):
        if isinstance(name, VNXStorageGroup):
            name = name.name

        return self._name_sg_map.get(name)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import storops.unity.resource.nas_server
from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class UnityNfsServer(UnityResource):
    @classmethod
    def create(cls, cli, nas_server, host_name=None, nfs_v4_enabled=True,
               is_secure_enabled=None, is_extended_credentials_enabled=None,
               kdc_type=None, kdc_username=None, kdc_password=None):
        clz = storops.unity.resource.nas_server.UnityNasServer
        nas_server = clz.get(cli, nas_server)

        resp = cli.post(
            cls().resource_class,
            nasServer=nas_server,
            hostName=host_name,
            nfsv4Enabled=nfs_v4_enabled,
            isSecureEnabled=is_secure_enabled,
            isExtendedCredentialsEnabled=is_extended_credentials_enabled,
            kdcType=kdc_type,
            kdcUsername=kdc_username,
            kdcPassword=kdc_password)

        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def delete(self, skip_kdc_unjoin=None, username=None,
               password=None, async_mode=False):
        resp = self._cli.delete(self.resource_class,
                                self.get_id(),
                                skipUnjoin=skip_kdc_unjoin,
                                kdcUsername=username,
                                kdcPassword=password,
                                async_mode=async_mode)
        resp.raise_if_err()
        return resp


class UnityNfsServerList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityNfsServer



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals, division

import inspect
import math
import os
import sys
from functools import reduce

from storops.lib.common import cache, all_not_none, allow_omit_parentheses
from storops.lib.metric import MetricConfigList, MetricConfigParser, \
    CalculatorMetaInfo

__author__ = 'Cedric Zhuang'


def get_counter(props):
    if isinstance(props, (list, tuple, set)):
        if len(props) != 1:
            raise ValueError('takes in one and only one property.')
        props = next(iter(props))
    return props


NaN = float('nan')


def isnan(value):
    return math.isnan(value)


def is_valid(value):
    return value is not None and not isnan(value)


def add(op1, op2):
    ret = 0.0
    if is_valid(op1):
        ret += op1
    if is_valid(op2):
        ret += op2
    return ret


def minus(op1, op2):
    if all_not_none(op1, op2):
        ret = op1 - op2
    else:
        ret = NaN
    return ret


def mul(op1, op2):
    if all_not_none(op1, op2):
        ret = op1 * op2
    else:
        ret = NaN
    return ret


def div(op1, op2):
    ret = NaN
    if all_not_none(op1, op2):
        if op1 == 0:
            ret = 0.0
        elif op2 != 0:
            ret = op1 / op2
    return ret


def round_60(value):
    """ round the number to the multiple of 60

    Say a random value is represented by: 60 * n + r
    n is an integer and r is an integer between 0 and 60.
    if r < 30, the result is 60 * n.
    otherwise, the result is 60 * (n + 1)

    The use of this function is that the counter refreshment on
    VNX is always 1 minute.  So the delta time between samples of
    counters must be the multiple of 60.
    :param value: the value to be rounded.
    :return: result
    """
    t = 60
    if value is not None:
        r = value % t
        if r > t / 2:
            ret = value + (t - r)
        else:
            ret = value - r
    else:
        ret = NaN
    return ret


@allow_omit_parentheses
def instance_stats_calculator(per_second=False):
    def decorator(func):
        def wrapper(prev, curr, obj, counters):
            prev_rsc = prev.get_rsc(obj)
            curr_rsc = curr.get_rsc(obj)

            if all_not_none(prev_rsc, curr_rsc):
                ret = func(prev_rsc, curr_rsc, counters)
                if per_second:
                    dt = round_60(curr.delta_seconds(prev))
                    ret = div(ret, dt)
            else:
                ret = NaN
            return ret

        return wrapper

    return decorator


@instance_stats_calculator
def stats_total(prev, curr, counters):
    return reduce(add, (getattr(curr, counter) for counter in counters), 0)


@instance_stats_calculator
def utilization(prev, curr, counters):
    """ calculate the utilization

    delta_busy = curr.busy - prev.busy
    delta_idle = curr.idle - prev.idle
    utilization = delta_busy / (delta_busy + delta_idle)

    :param prev: previous resource
    :param curr: current resource
    :param counters: list of two, busy ticks and idle ticks
    :return: value, NaN if invalid.
    """
    busy_prop, idle_prop = counters

    pb = getattr(prev, busy_prop)
    pi = getattr(prev, idle_prop)

    cb = getattr(curr, busy_prop)
    ci = getattr(curr, idle_prop)

    db = minus(cb, pb)
    di = minus(ci, pi)

    return mul(div(db, add(db, di)), 100)


def aggregated_sum(rsc_list, counters):
    counter = get_counter(counters)
    return reduce(add, getattr(rsc_list, counter), 0.0)


@instance_stats_calculator(per_second=True)
def delta_ps(prev, curr, counters):
    """ calculate the delta per second of one counter

    formula: (curr - prev) / delta_time
    :param prev: previous resource
    :param curr: current resource
    :param counters: the counter to do delta and per second, one only
    :return: value, NaN if invalid.
    """
    counter = get_counter(counters)

    pv = getattr(prev, counter)
    cv = getattr(curr, counter)
    return minus(cv, pv)


@instance_stats_calculator
def io_size_kb(prev, curr, counters):
    """ calculate the io size based on bandwidth and throughput

    formula: average_io_size = bandwidth / throughput
    :param prev: prev resource, not used
    :param curr: current resource
    :param counters: two stats, bandwidth in MB and throughput count
    :return: value, NaN if invalid
    """
    bw_stats, io_stats = counters
    size_mb = div(getattr(curr, bw_stats), getattr(curr, io_stats))
    return mul(size_mb, 1024)


def self_io_size_kb(obj, counters):
    left, right = counters
    return mul(div(getattr(obj, left), getattr(obj, right)), 1024)


def block_to_mbps(prev, curr, obj, counters):
    return delta_ps(prev, curr, obj, counters) * 512 / 2 ** 20


def kb_to_mbps(prev, curr, obj, counters):
    return delta_ps(prev, curr, obj, counters) / 2 ** 10


@cache
def _module_functions():
    return dict(inspect.getmembers(sys.modules[__name__]))


class VNXMetricConfig(object):
    def __init__(self, config):
        self.name = config.get('name')
        self.calculator = self._get_calculator(config)
        self.counters = config.get('counters')
        self.aggregated_from = config.get('aggregated_from')

    @staticmethod
    def _get_calculator(config):
        calc_name = config.get('calculator')
        if not calc_name:
            ret = delta_ps
        else:
            ret = _module_functions().get(calc_name)
        return ret

    def is_aggregated_stats(self):
        return self.aggregated_from is not None


class VNXMetricConfigList(MetricConfigList):
    @classmethod
    def init_metric_config(cls, raw_config):
        return VNXMetricConfig(raw_config)


class VNXMetricConfigParser(MetricConfigParser):
    @classmethod
    def get_config(cls, name):
        name = cls._get_clz_name(name)
        return VNXMetricConfigList(cls._read_configs().get(name))

    @classmethod
    def get_folder(cls):
        return os.path.dirname(inspect.getfile(cls))


class VNXCalculatorMetaInfo(CalculatorMetaInfo):
    def get_config_parser(self):
        return VNXMetricConfigParser()

    def get_metric_value(self, clz, metric_name, cli, obj=None):
        if not hasattr(cli, 'curr_counter'):
            raise ValueError('cli should has "curr_counter" attribute.')
        if not hasattr(cli, 'prev_counter'):
            raise ValueError('cli should has "prev_counter" attribute.')

        config = self.get_config(clz).get_metric_config(metric_name)
        if config.is_aggregated_stats():
            ret = self._get_aggregated_stats(config, obj)
        else:
            ret = self._get_calculated_stats(cli, config, obj)
        return ret

    @staticmethod
    def _get_calculated_stats(cli, config, obj):
        prev = cli.prev_counter
        curr = cli.curr_counter
        if all_not_none(prev, curr):
            ret = config.calculator(prev, curr, obj, config.counters)
        else:
            ret = NaN
        return ret

    @staticmethod
    def _get_aggregated_stats(config, obj):
        if 'self' in config.aggregated_from:
            ret = config.calculator(obj, config.counters)
        else:
            rsc_list = getattr(obj, config.aggregated_from)
            if rsc_list is None:
                rsc_list = []
            ret = config.calculator(rsc_list, config.counters)
        return ret


calculators = VNXCalculatorMetaInfo()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals
import unittest

from hamcrest import assert_that, greater_than_or_equal_to, raises
from hamcrest import equal_to

from storops_test.vnx.nas_mock import t_nas, patch_nas
from storops.vnx.enums import VNXShareType
from storops.exception import VNXBackendError, VNXInvalidMoverID, \
    VNXMoverInterfaceNotAttachedError, VNXMoverInterfaceNotExistsError
from storops.vnx.resource.vdm import VNXVdm

__author__ = 'Jay Xu'


class VNXVdmTest(unittest.TestCase):
    @patch_nas
    def test_get_all(self):
        vdm_list = VNXVdm.get(t_nas())
        assert_that(len(vdm_list), greater_than_or_equal_to(1))
        dm = next(dm for dm in vdm_list if dm.vdm_id == 2)
        self.verify_vdm_2(dm)

    @patch_nas
    def test_get_by_id_invalid(self):
        dm = VNXVdm.get(vdm_id=1, cli=t_nas())
        assert_that(dm.existed, equal_to(False))

    @patch_nas
    def test_get_by_id_2(self):
        dm = VNXVdm(vdm_id=2, cli=t_nas())
        self.verify_vdm_2(dm)

    @patch_nas
    def test_get_by_name(self):
        dm = VNXVdm.get(name='VDM_ESA', cli=t_nas())
        self.verify_vdm_2(dm)

    @patch_nas
    def test_get_by_name_not_found(self):
        dm = VNXVdm(name='not_found', cli=t_nas())
        assert_that(dm.existed, equal_to(False))

    @staticmethod
    def verify_vdm_2(dm):
        assert_that(dm.root_fs_id, equal_to(199))
        assert_that(dm.mover_id, equal_to(1))
        assert_that(dm.name, equal_to('VDM_ESA'))
        assert_that(dm.existed, equal_to(True))
        assert_that(dm.vdm_id, equal_to(2))
        assert_that(dm.state, equal_to('loaded'))
        assert_that(dm.status, equal_to('ok'))
        assert_that(dm.is_vdm, equal_to(True))

    @patch_nas
    def test_create_vdm_invalid_mover_id(self):
        def f():
            VNXVdm.create(t_nas(), 3, 'myVdm')

        assert_that(f, raises(VNXInvalidMoverID))

    @patch_nas
    def test_create_vdm(self):
        dm = VNXVdm.create(t_nas(), 2, 'myVdm')
        assert_that(dm.name, equal_to('myVdm'))
        assert_that(dm.vdm_id, equal_to(3))
        assert_that(dm.mover_id, equal_to(2))
        assert_that(dm.root_fs_id, equal_to(245))

    @patch_nas
    def test_delete_vdm(self):
        dm = VNXVdm(vdm_id=3, cli=t_nas())
        resp = dm.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_delete_vdm_not_found(self):
        def f():
            dm = VNXVdm(vdm_id=5, cli=t_nas())
            dm.delete()

        assert_that(f, raises(VNXBackendError, 'not found'))

    @patch_nas
    def test_attach_interface_success(self):
        dm = VNXVdm(name='myvdm', cli=t_nas())
        dm.attach_nfs_interface('1.1.1.1-0')

    @patch_nas
    def test_attach_interface_not_found(self):
        def f():
            dm = VNXVdm(name='myvdm', cli=t_nas())
            dm.attach_nfs_interface('1.1.1.2-0')

        assert_that(f, raises(VNXMoverInterfaceNotExistsError, 'not exist'))

    @patch_nas
    def test_detach_interface_success(self):
        dm = VNXVdm(name='myvdm', cli=t_nas())
        dm.detach_nfs_interface('1.1.1.1-0')

    @patch_nas
    def test_detach_interface_not_found(self):
        def f():
            dm = VNXVdm(name='myvdm', cli=t_nas())
            dm.detach_nfs_interface('1.1.1.2-0')

        assert_that(f, raises(VNXMoverInterfaceNotExistsError, 'not exist'))

    @patch_nas
    def test_detach_interface_not_attached(self):
        def f():
            dm = VNXVdm(name='myvdm', cli=t_nas())
            dm.detach_nfs_interface('1.1.1.3-0')

        assert_that(f, raises(VNXMoverInterfaceNotAttachedError, 'attached'))

    @patch_nas
    def test_get_interfaces(self):
        dm = VNXVdm(name='VDM_ESA', cli=t_nas())
        ifs = dm.get_interfaces()
        assert_that(len(ifs), equal_to(1))
        interface = ifs[0]
        assert_that(interface.name, equal_to('10-110-24-195'))
        assert_that(interface.share_type, equal_to(VNXShareType.NFS))
